"""."""

import datetime
import json
import re
import time

import lxml.html
import networkx as nx
import pyscreenshot
import pyvirtualdisplay
import requests
import selenium
from selenium import webdriver
from selenium.webdriver.common.by import By
from selenium.webdriver.common.action_chains import ActionChains
from selenium.webdriver.support import expected_conditions
from selenium.webdriver.support.ui import WebDriverWait

from understory import mf
from understory import solarized
from understory import uri
from understory import web

__all__ = ["post", "get", "put", "delete",
           "parse", "browser", "cache", "discover_link"]

displays = []
browsers = []

tor_proxies = {"http": "socks5h://localhost:9050",
               "https": "socks5h://localhost:9050"}


def discover_link(target, name):
    """
    Fetch target and discover link `name` in HTML rels or HTTP Link header.

    Skip HEAD request; use GET and attempt to cache.

    """
    try:
        response = web.tx.cache[target]
    except AttributeError:
        response = get(target)
    try:
        link = _get_header_link(response.headers, name)[0]
        if link.startswith("/"):
            endpoint = uri.parse(target)
            endpoint.path = link
        else:
            endpoint = link
    except IndexError:
        try:
            endpoint = uri.parse(response.mf2json["rels"].get(name, [])[0])
        except IndexError:
            endpoint = None
    return endpoint


def _get_header_link(headers: dict, search_rel: str):
    try:
        header = headers["Link"]
    except KeyError:
        return []
    links = []
    for link in header.split(","):
        resource, _, rel = link.partition(";")
        match = re.match("""rel=['"](.+)['"]""", rel.strip())
        if match and match.groups()[0] == search_rel:
            links.append(resource.strip(" <>"))
    return links


def download(url, filepath, chunk_size=1024):
    """Download url to filepath."""
    response = request("GET", url, stream=True)
    with filepath.open("wb") as fp:
        for chunk in response.iter_content(chunk_size=chunk_size):
            if chunk:
                fp.write(chunk)
    return response


def request(method, url, **kwargs):
    """
    Return the response to dereferencing given `url` using given `method`.

    Attempts to use HTTPS when accessing non-onion domains. Proxies
    through Tor when accessing onion services. Optionally pass typical
    `requests.Request` arguments as `kwargs`.

    """
    url = uri.parse(url)
    preferred = "https"
    if url.suffix == "onion":
        kwargs["proxies"] = tor_proxies
        preferred = "http"
    try:
        response = requests.request(method, f"{preferred}://{url.minimized}",
                                    verify=False, **kwargs)
    except (requests.exceptions.SSLError, requests.exceptions.ConnectionError):
        if url.suffix != "onion":
            try:
                response = requests.request(method, f"http://{url.minimized}",
                                            **kwargs)
            except (requests.exceptions.SSLError,
                    requests.exceptions.ConnectionError):
                raise RequestFailed()
    return response


class RequestFailed(Exception):
    """A request has failed."""


def post(url, **kwargs):
    """Post to the web."""
    return Transaction(url, "post", **kwargs)


def get(url, **kwargs):
    """Get from the web."""
    return Transaction(url, "get", **kwargs)


def put(url, **kwargs):
    """Put to the web."""
    return Transaction(url, "put", **kwargs)


def delete(url, **kwargs):
    """Delete from the web."""
    return Transaction(url, "delete", **kwargs)


class Cache:

    def __init__(self, domain=None, db=None):
        self.cache = {}
        if domain:
            self.domain = domain
        if db:
            self.db = db
            self.db.define(cache="""url TEXT UNIQUE, html TEXT,
                                    headers JSON, data JSON""")

    def format_url(self, url):
        try:
            url = f"{self.domain}/{url}"
        except AttributeError:
            url = url
        return url

    def add(self, url):
        url = self.format_url(url)
        resource = get(url)
        self.cache[url] = resource
        try:
            self.db.insert("cache", url=url, html=resource.text,
                           headers=dict(resource.headers),
                           data=resource.mf2json.data)
        except self.db.IntegrityError:
            self.db.update("cache", html=resource.text,
                           headers=dict(resource.headers),
                           data=resource.mf2json.data, where="url = ?",
                           vals=[url])
        except AttributeError:
            pass
        return resource

    def __getitem__(self, resource_url):
        resource_url = str(resource_url)
        try:
            resource = self.cache[resource_url]
        except KeyError:
            try:
                url = self.format_url(resource_url)
                resource_data = self.db.select("cache", where="url = ?",
                                               vals=[url])[0]
                resource = Transaction(url, fetch=False)
                resource.headers = resource_data["headers"]  # TODO case-insen
                resource.text = resource_data["html"]
            except (AttributeError, IndexError):
                self.add(resource_url)
                resource = self.cache[resource_url]
        return resource

    @property
    def graph(self):
        network = nx.DiGraph()
        for url, resource in self.cache.items():
            # print(resource.links)
            network.add_node(url)
        return nx.draw(network, with_labels=True)


cache = Cache


class Transaction:
    """."""

    def __init__(self, url, method="get", fetch=True, **kwargs):
        self.url = str(uri.parse(str(url)))
        if fetch:
            # XXX handler = getattr(requests, method)
            # XXX self.response = handler(apply_dns(self.url), **kwargs)
            self.response = request(method, self.url, **kwargs)
            self.text = self.response.text
            self.headers = self.response.headers

    @property
    def location(self):
        location = self.headers["location"]
        if location.startswith("/"):
            new_url = uri.parse(self.response.url)
            new_url.path = location
            location = str(new_url)
        return location

    @property
    def links(self):
        return self.response.links

    @property
    def dom(self):
        return parse(self.text)

    @property
    def json(self):
        return json.loads(self.text)

    @property
    def mf2json(self):
        return Semantics(mf.parse(self.text, self.url))

    @property
    def card(self):
        return Semantics(mf.representative_hcard(self.mf2json.data,
                                                 source_url=self.url))

    @property
    def feed(self):
        return Semantics(mf.interpret_feed(self.mf2json.data,
                                           source_url=self.url))

    @property
    def entry(self):
        return Semantics(mf.interpret_entry(self.mf2json.data,
                                            source_url=self.url))

    @property
    def event(self):
        return Semantics(mf.interpret_event(self.mf2json.data,
                                            source_url=self.url))

    def mention(self, *target_urls):
        return Semantics(mf.interpret_comment(self.mf2json.data,
                                              self.url, target_urls))

    # @property
    # def jf2(self):
    #     return Semantics(mf.interpret_feed(self.mf2json.data,
    #                                        source_url=self.url))


class Semantics:

    def __init__(self, data):
        self.data = data

    def __getitem__(self, item):
        return self.data[item]

    def __repr__(self):
        return JSONEncoder(indent=2).encode(self.data)  # , indent=2)
        # XXX return json.dumps(self.data, indent=2)

    def _repr_html_(self):
        return solarized.highlight(JSONEncoder(indent=2).encode(self.data),
                                   ".json")
        # XXX return solarized.highlight(json.dumps(self.data, indent=2),
        # ".json")

    def __bool__(self):
        return bool(self.data)


class JSONEncoder(json.JSONEncoder):

    def default(self, obj):
        if isinstance(obj, uri.URI):
            return str(obj)
        if isinstance(obj, (datetime.date, datetime.datetime)):
            return obj.isoformat()
        return json.JSONEncoder.default(self, obj)


def parse(html):
    """Return a document object for given html."""
    return Document(html)


def apply_dns(url):
    if url.startswith("/"):
        return url
    url = uri.parse(url)
    if url.host == "alice.example":
        url = str(url).replace("http://alice.example",
                               "http://127.0.0.1:8080")
    elif url.host == "bob.example":
        url = str(url).replace("http://bob.example",
                               "http://127.0.0.1:8081")
    elif url.host == "hello.example":
        url = str(url).replace("http://hello.example",
                               "http://127.0.0.1:8082")
    else:
        url = str(url)
    return url


def unapply_dns(url):
    url = uri.parse(url)
    if url.host == "127.0.0.1":
        if url.port == 8080:
            url = str(url).replace("http://127.0.0.1:8080",
                                   "http://alice.example")
        elif url.port == 8081:
            url = str(url).replace("http://127.0.0.1:8081",
                                   "http://bob.example")
        elif url.port == 8082:
            url = str(url).replace("http://127.0.0.1:8082",
                                   "http://hello.example")
    else:
        url = str(url)
    return url


class Document:

    # TODO with html as dom: -- manipulate dom -- on exit html is modified

    def __init__(self, html):
        self.doc = lxml.html.fromstring(str(html))

    def select(self, selector):
        els = []
        for el in self.doc.cssselect(selector):
            els.append(Element(el))
        return els

    @property
    def children(self):
        return self.doc.getchildren()

    @property
    def html(self):
        return lxml.html.tostring(self.doc).decode()


class Element:

    def __init__(self, element):
        self.element = element

    def append(self, *html):
        for _html in html:
            self.element.append(_make_element(_html))

    def select(self, selector):
        els = []
        for el in self.element.cssselect(selector):
            els.append(Element(el))
        return els

    def replace(self, html):
        self.element.getparent().replace(self.element, _make_element(html))

    # TODO automatically add only if an: a or link
    @property
    def href(self):
        return self.element.attrib["href"]

    @property
    def text(self):
        return self.element.text_content()


def _make_element(html):
    el = lxml.html.fromstring(f"<DOUGIE>{html}</DOUGIE>")
    return el.cssselect("DOUGIE")[0].getchildren()[0]


class Browser:
    """Firefox via Selenium."""

    By = By
    EC = expected_conditions

    def __init__(self, name=None, width=1024, height=768):
        if not len(displays):
            display = pyvirtualdisplay.Display(visible=False, size=(2048, 768))
            display.start()
            displays.append(display)
        profile = webdriver.FirefoxProfile()
        profile.add_extension(extension="/home/gaea/canopy/var/identities/"
                                        "6c189616-4fe1-4f3f-84dc-c4a13ee9b155/"
                                        "asteria/asteria-dev.xpi")
        binary = "/home/gaea/firefox/firefox-bin"
        self.browser = webdriver.Firefox(firefox_profile=profile,
                                         firefox_binary=binary)
        count = len(browsers)
        browsers.append(self)
        self._top = 0
        self._left = count * 1024
        self._width = width
        self._height = height
        self._update_window()
        self.name = name
        self.shot_id = 0

    @property
    def width(self):
        return self._width

    @width.setter
    def width(self, value):
        self._width = value
        self._update_window()

    @property
    def height(self):
        return self._height

    @height.setter
    def height(self, value):
        self._height = value
        self._update_window()

    def _update_window(self):
        self.browser.set_window_rect(self._left, self._top,
                                     self._width, self._height)

    def go(self, *args, wait=0):
        if len(args) == 1:
            url = args[0]
        elif len(args) == 2:
            url = "/".join(args)
        url = apply_dns(url)
        self.browser.get(url)
        if wait:
            time.sleep(wait)
        return self
        # XXX self.browser.get(str(uri.parse(url)))

    def wait(self, *conditions, duration=20):
        for condition in conditions:
            time.sleep(.1)
            wait = WebDriverWait(self.browser, duration)
            wait.until(condition)

    def wait_until_url_contains(self, url):
        self.wait(self.EC.url_contains(apply_dns(url)))

    def select(self, selector):
        return self.browser.find_elements_by_css_selector(selector)

    def select_first(self, selector):
        return self.browser.find_element_by_css_selector(selector)

    def action(self):
        return ActionChains(self.browser)

    def shot(self, path):
        # TODO take in pieces & stitch together -- using way too much memory
        # self._height = self.browser.execute_script("return document.body."
        #                                            "scrollHeight;") + 100
        # self._update_window()
        self.browser.get_screenshot_as_file(str(path) + ".png")

    # XXX def shot_url(self):
    # XXX     base64 = self.browser.get_screenshot_as_base64()
    # XXX     return f"data:image/png;BASE64,{base64}"

    def shot_url(self):
        # XXX grab = pyscreenshot.grab(bbox=(0, 0, 920, 920)).tobytes()
        # XXX base64png = b"".join(base64.encodebytes(grab).splitlines())
        self.shot_id += 1
        filename = f"{self.name}-{self.shot_id}.png"
        placement = browsers.index(self)
        coords = (1024 * placement, 0,
                  1024 * (placement + 1), 768)
        # import sh
        # sh.Command("import")("-screen", "-window", "root", filename)
        # time.sleep(2)
        # sh.Command("import")("-window", "root", filename)
        # sh.convert(sh.xwd("-root", "-screen"), "xwd:-", f"png:{filename}")
        pyscreenshot.grab(bbox=coords).save(filename)
        return f"/IndieWeb/{filename}"

    def quit(self):
        try:
            self.browser.quit()
        except selenium.common.exceptions.WebDriverException:
            pass
        if displays:
            try:
                displays[0].stop()
            except KeyError:  # raising during multi-user testing
                pass

    def __getattr__(self, attr):
        return getattr(self.browser, attr)

    def _repr_html_(self):
        return f"<img class=screen src={self.shot_url()}>"
        # url = unapply_dns(self.current_url)
        # site_character = url.partition("//")[2].partition(".")[0]
        # TODO FIXME remove hard-coded IndieWeb..
        # return (f"<div class=screenshot>"
        #         f"<div class=browser><small>{self.name}'s "
        #         f"Browser</small></div>"
        #         f"<div class=tab><img src=/IndieWeb/{site_character}16.png>"
        #         f" {self.title}</div>"
        #         f"<div class=address><small><code>{url}</code></small></div>"
        #         f"<img class=screen src={self.shot_url()}></div>")


browser = Browser


# def shot(name, description):  # XXX , *browsers):
#     test_case = inspect.stack()[1].function
#     global shot_counter
#     shot_id = "{:03}".format(shot_counter)
#     dashed_name = name.replace(" ", "-")
#     for user, browser in sorted(browsers.items()):
#         shot_filename = "{}-{}-{}.png".format(shot_id, user, dashed_name)
#         height = browser.execute_script("return document.body.scrollHeight;")
#         browser.set_window_size(browser_width, height + 100)
#         browser.get_screenshot_as_file(str(build_dir / "features" /
#                                            shot_filename))
#     features.append((test_case, shot_id, dashed_name, name, description))
#     # XXX , shot_filename, [user for u in browsers.keys()]))
#     shot_counter += 1



"""

"""

# XXX import configparser
# XXX import os
import re

import httpagentparser
import mimeparse

from . import util

__all__ = ["Accept", "AcceptCharset", "AcceptEncoding", "AcceptLanguage",
           "Authorization", "Cookie", "Expect", "From", "Host", "IfMatch",
           "IfModifiedSince", "IfNoneMatch", "IfRange", "IfUnmodifiedSince",
           "KeepAlive", "MaxForwards", "ProxyAuthorization", "Range",
           "Referer", "TE", "UserAgent", "XBehavioralAdOptOut",
           "XDoNotTrack", "XForwardedFor", "XRequestedWith"]


class Request(util.Header):

    """"""


class _Accept(Request):

    """"""

    def parse(self):
        acceptables = []
        noq = 1
        for acceptable in self.header.split(","):
            value, _, params = acceptable.strip().lower().partition(";")
            parameters = {}
            if params.strip():
                for param in params.split(";"):
                    k, _, v = param.partition("=")
                    parameters[k.strip()] = v.strip()
            noq -= .00001
            q = float(parameters.pop("q", noq))
            acceptables.append((q, (value.strip(), parameters)))
        self.acceptables = [self.type(a[0], _q, **a[1]) for _q, a in
                            list(reversed(sorted(acceptables)))]

    class _Acceptable:

        """"""

        def __init__(self, value, quality, **params):
            self.value = value.lower()
            self.quality = quality
            self.params = {k.lower(): v.lower() for k, v in params.items()}
            self.parse()

        def __repr__(self):
            return repr(self.value)


class Accept(_Accept):

    """
    The "Accept" request-header field can be used by user agents to specify
    response media types that are acceptable. Accept header fields can be
    used to indicate that the request is specifically limited to a small
    set of desired types, as in the case of a request for an in-line image.

        >>> header = '''text/*;q=0.3, text/html;q=0.7, text/html;level=1,
        ...             text/html;level=2;q=0.4, */*;q=0.5'''
        >>> accept = Accept(header)
        >>> accept.acceptables
        ['text/html', 'text/html', '*/*', 'text/html', 'text/*']
        >>> type(accept.acceptables[0])
        <class 'web.headers.request.Accept.Media'>
        >>> accept.best_match(["image/png", "text/plain", "text/html"])
        'text/html'

    """

    def best_match(self, supported):
        return mimeparse.best_match(supported, self.header)

    class Media(_Accept._Acceptable):

        """"""

        def parse(self):
            self.canonical = self.value

    type = Media

    _meta = "TYPE"


class AcceptCharset(_Accept):

    """
    The "Accept-Charset" request-header field can be used by user agents
    to indicate what response character sets are acceptable. This field
    allows clients capable of understanding more comprehensive or
    special-purpose character sets to signal that capability to a server
    which is capable of representing documents in those character sets.

    """

    class Charset(_Accept._Acceptable):

        """"""

        def parse(self):
            self.canonical = self.value

    type = Charset

    _meta = "CHARSET"


class AcceptEncoding(_Accept):

    """"""

    # def best_match(self, supported):
    #   acceptable = [p.lower().strip() for p in self.header.split(",")]
    #   for encoding in supported:
    #     if encoding in acceptable:
    #       return encoding
    #   raise http.NotAcceptable("asdf")

    class Encoding(_Accept._Acceptable):

        """"""

        def parse(self):
            self.canonical = self.value

    type = Encoding

    _meta = "ENCODING"


class AcceptLanguage(_Accept):

    """"""

    class Language(_Accept._Acceptable):

        """"""

        def parse(self):
            self.canonical = self.value

    type = Language

    _meta = "LANGUAGE"


class Authorization(Request):

    """"""

    _meta = ""


class Cookie(Request):

    """"""

    def parse(self):
        self.morsels = {}
        for morsel in self.header.split(";"):
            k, _, v = morsel.lstrip().partition("=")
            self.morsels[k] = v

    def get(self, key, default=None):
        return self.morsels.get(key, default)


class Expect(Request):

    """"""


class From(Request):

    """"""


class Host(Request):

    """"""

    # TODO support for IP addresses and lazy determination

    def parse(self):
        self.header = self.header.lower()
        self.name, _, port = self.header.partition(":")
        if port is None:
            port = "80"
        self.port = port

    @property
    def is_hostname(self):
        """"""
        if len(self.name) > 255:
            return False
        allowed = re.compile(r"^(?!-)[a-z\d-]{1,63}(?<!-)$")  # TODO raw ok?
        return all(allowed.match(l) for l in self.name.strip(".").split("."))


class IfMatch(Request):

    """"""


class IfModifiedSince(Request):

    """"""


class IfNoneMatch(Request):

    """"""


class IfRange(Request):

    """"""


class IfUnmodifiedSince(Request):

    """"""


class KeepAlive(Request):

    """"""


class MaxForwards(Request):

    """"""


class ProxyAuthorization(Request):

    """"""


class Range(Request):

    """"""


class Referer(Request):

    """"""


class TE(Request):

    """"""


class UserAgent(Request):

    """"""

    def parse(self):
        self.features = httpagentparser.simple_detect(self.header)

    # _agents = {}
    # _re_agents = {}
    # _defaults = None

    # def parse(self):
    #     if not self._agents:
    #         self._initialize_browscap()
    #     possibles = list(name for pattern, name in self._re_agents.items()
    #                      if pattern.match(self.header))
    #     self.features = self._defaults
    #     if possibles:
    #         self.features = self._agents[max(possibles,
    #                                          key=lambda n: len(n))]

    # @property
    # def is_js_compatible(self):
    #     return self.features["javascript"] == "true"

    # @classmethod
    # def _initialize_browscap(cls, path=None):
    #     if path is None:
    #         path = cls._get_filename()
    #     browscap = configparser.ConfigParser()
    #     if not browscap.read(path):
    #         return
    #     browscap.remove_section("GJK_Browscap_Version")
    #     defaults = dict(browscap.items("DefaultProperties"))
    #     browscap.remove_section("DefaultProperties")
    #     browscap.remove_section("*")  # TODO fall back to default browser
    #     families = {}
    #     for sect in browscap.sections():
    #         if browscap.get(sect, "Parent") == "DefaultProperties":
    #             families[sect] = dict(defaults, **dict(browscap.items(sect)))
    #             browscap.remove_section(sect)
    #     for sect in browscap.sections():
    #         parent = browscap.get(sect, "Parent")
    #         cls._agents[sect] = dict(families[parent],
    #                                  **dict(browscap.items(sect)))
    #         pattern = sect
    #         for unsafe in "().-":
    #             pattern = pattern.replace(unsafe, "\\" + unsafe)
    #         pattern = pattern.replace("?", ".").replace("*", ".*?")
    #         cls._re_agents[re.compile(r"^" + pattern + r"$")] = sect
    #     cls._defaults = defaults

    # @staticmethod
    # def _get_filename():
    #     return os.path.join(os.path.dirname(__file__), "browscap.ini")

    # @staticmethod
    # def _update_browscap():
    #     import httplib2
    #     agent = httplib2.Http()
    #     print("Updating browscap file..")

    #     uri = "http://browsers.garykeith.com/stream.asp?BrowsCapINI"
    #     print("Downloading from:", uri)
    #     content = agent.request(uri)[1]

    #     path = UserAgent._get_filename()
    #     print("Saving to:", path)
    #     with open(path, "w") as file:
    #         file.write(content)

    #     print("Success.")

# XXX UserAgent._initialize_browscap()


class XBehavioralAdOptOut(Request):

    """"""

    def __repr__(self):
        return repr(bool(self.header))


class XDoNotTrack(Request):

    """"""

    def __repr__(self):
        return repr(bool(self.header))


class XForwardedFor(Request):

    """"""


class XRequestedWith(Request):

    """"""

    @property
    def ajax(self):
        return self.header == "XMLHttpRequest"

    def __repr__(self):
        return repr(self.header)



"""

"""

import calendar
import datetime
import email

from .util import Header

__all__ = ["Allow", "ContentEncoding", "ContentLanguage", "ContentLength",
           "ContentLocation", "ContentMD5", "ContentRange", "ContentType",
           "Expires", "LastModified"]


def get_timestamp(dt, fmt="rfc822"):
    """
    return a string of datetime `dt` in common log format

        >>> get_timestamp(datetime.datetime(1970, 1, 1, 1, 1, 1))
        'Thu, 01 Jan 1970 01:01:01 GMT'

    """
    if dt is None:
        dt = datetime.UTC.localize(datetime.now())
    fmts = {"rfc822": "%a, %d %b %Y %H:%M:%S GMT",
            "ncsa": "%d/%b/%Y:%H:%M:%S %z"}
    return dt.strftime(fmts[fmt])


def parse_datetime(s):
    """
    parses a date/time stamp taken from an HTTP header

    """
    return datetime(*email.utils.parsedate(s)[:6])


class Entity(Header):

    """"""


class Allow(Entity):

    """"""


class ContentEncoding(Entity):

    """"""


class ContentLanguage(Entity):

    """"""


class ContentLength(Entity):

    """"""


class ContentLocation(Entity):

    """"""


class ContentMD5(Entity):

    """"""


class ContentRange(Entity):

    """"""


class ContentType(Entity):

    """"""

    @property
    def content_type(self):
        _content_type = str(self).partition(";")[0]
        if not _content_type:
            _content_type = "text/plain"
        return _content_type


class Expires(Entity):

    """
    formats an `Expires` header for `delta` from now

    `delta` is a `datetime.timedelta` object or a number of seconds.

    """

    def __init__(self, when):
        # FIXME either seconds a la TTL or datetime according to W3C
        now = datetime.datetime.utcnow()
        if isinstance(when, str):
            try:
                when = int(when)
            except ValueError:
                then = parse_datetime(when)
                when = (calendar.timegm(then.utctimetuple()) -
                        calendar.timegm(now.utctimetuple()))
        if isinstance(when, int):
            when = datetime.timedelta(seconds=when)
        dt = now + when
        self.header = get_timestamp(dt=dt)


class LastModified(Entity):

    """"""

    # def __init__(self, when):
    #     self.header = util.format_datetime(when)



"""Micropub server app."""

import pathlib

import pendulum
import sh
import vobject

from understory import web
from understory.web import tx

from .. import webmention


server = web.application("MicropubServer", mount_prefix="pub", db=False,
                         channel=r".+", entry=r".+", nickname=r"[A-Za-z0-9-]+",
                         filename=rf"{web.nb60_re}{{4}}.\w{{1,10}}")
templates = web.templates(__name__)

# TODO supported_types = {"RSVP": ["in-reply-to", "rsvp"]}


def send(properties, endpoint=None, h="entry", token=None):
    """Send a Micropub request to a Micropub server."""
    # TODO FIXME what's in the session?
    if endpoint is None:
        endpoint = tx.user.session["micropub_endpoint"]
    response = web.post(endpoint, headers={"Authorization": f"Bearer {token}"},
                        json={"type": [f"h-{h}"], "properties": properties})
    return response.location, response.links


def define_table():
    """Define the `resources`, `media` and `syndication` tables."""
    tx.db.define(resources="""resource JSON, version TEXT UNIQUE,
                              parents JSON""",
                 # TODO updates="""""",
                 media="""mid TEXT, sha256 TEXT UNIQUE, size INTEGER""",
                 syndication="""destination JSON NOT NULL""")


def wrap_server(handler, app):
    """Ensure server links are in head of root document."""
    define_table()
    tx.pub = LocalClient()
    try:
        tx.host.owner = tx.pub.get_card("me")
    except IndexError:
        tx.host.owner = {}
    yield
    if tx.request.uri.path == "" and tx.response.body:
        doc = web.parse(tx.response.body)
        try:
            head = doc.select("head")[0]
        except IndexError:
            pass
        else:
            head.append("<link rel=micropub href=/pub>")
            tx.response.body = doc.html
        web.header("Link", '</pub>; rel="micropub"', add=True)


def route_unrouted(handler, app):
    """Handle channels."""
    for channel in tx.pub.get_channels():
        if channel["resource"]["url"][0] == f"/{tx.request.uri.path}":
            posts = tx.pub.get_posts_by_channel(channel["resource"]["uid"][0])
            web.header("Content-Type", "text/html")
            raise web.OK(templates.channel(channel, posts))
    yield


def discover_post_type(properties):
    """Return the discovered post type."""
    if "photo" in properties:
        post_type = "photo"
    elif "bookmark-of" in properties:
        post_type = "bookmark"
    elif "like-of" in properties:
        post_type = "like"
    elif "follow-of" in properties:
        post_type = "follow"
    elif "identification-of" in properties:
        post_type = "identification"
    elif "rsvp" in properties:
        post_type = "rsvp"
    elif "in-reply-to" in properties:
        post_type = "reply"
    elif "name" in properties:
        post_type = "article"
    else:
        post_type = "note"
    return post_type


def generate_vcard(nickname):
    """"""
    card = tx.pub.get_card(nickname)
    vcard = vobject.vCard()
    vcard.add("prodid").value = "-//Canopy//understory 0.0.0//EN"
    vcard.add("uid").value = card["uid"][0]
    vcard.add("fn").value = card["name"][0]
    return vcard.serialize()

    # TODO # TODO if identity["type"] == "identity":
    # TODO n = card.add("n")
    # TODO names = {}
    # TODO for name_type in ("prefix", "given", "additional",
    #                        "family", "suffix"):
    # TODO     if identity[name_type]:
    # TODO         names[name_type] = identity[name_type].split(";")
    # TODO n.value = vobject.vcard.Name(**names)
    # TODO # TODO else:
    # TODO # TODO     card.add("n")
    # TODO # TODO     card.add("org").value = [identity["name"]]

    # TODO # TODO card.add("nickname").value = identity["name"]
    # TODO card.add("sort_string").value = identity["sort_string"]

    # TODO for number, types in identity["telephones"]:
    # TODO     entry = card.add("tel")
    # TODO     entry.value = number
    # TODO     if types:
    # TODO         entry.params["TYPE"] = types

    # TODO for url, types in identity["websites"]:
    # TODO     entry = card.add("url")
    # TODO     entry.value = url
    # TODO     if types:
    # TODO         entry.params["TYPE"] = types

    # TODO try:
    # TODO     photo_id = identity["photos"][0]
    # TODO except IndexError:
    # TODO     pass
    # TODO else:
    # TODO     photo_data = \
    # TODO         canopy.branches["images"].photos.get_photo_data(id=photo_id)
    # TODO     photo = card.add("photo")
    # TODO     photo.value = photo_data
    # TODO     photo.encoding_param = "b"
    # TODO     photo.type_param = "JPEG"

    # item_index = 0
    # for vals in card.contents.values():
    #     for val in vals:
    #         if val.group:
    #             item_index = int(val.group[4:])

    # for related, types in get_relationships(identity["id"]):
    #     uri = "https://{}/identities/{}/{}.vcf".format(tx.host.name,
    #                                                related["identifier"],
    #                                                related["slug"])
    #     rel = card.add("related")
    #     rel.value = uri
    #     rel.params["TYPE"] = types
    #     for type in types:
    #         group_name = "item{}".format(item_index)
    #         rel_name = card.add("x-abrelatednames")
    #         rel_name.value = related["name"]
    #         rel_name.group = group_name
    #         rel_uri = card.add("x-aburi")
    #         rel_uri.value = uri
    #         rel_uri.group = group_name
    #         rel_type = card.add("x-ablabel")
    #         rel_type.value = "_$!<{}>!$_".format(type)
    #         rel_type.group = group_name
    #         item_index += 1


class LocalClient:
    """A localized interface to the endpoint's backend."""

    def create(self, resource_type, **resource):
        """Create a resource and return its permalink."""
        for k, v in resource.items():
            if not isinstance(v, list):
                resource[k] = [v]
        if "published" in resource:
            resource["published"][0]["datetime"] = \
                pendulum.from_format(resource["published"][0]["datetime"],
                                     "YYYY-MM-DDTHH:mm:ssZ")
            published = resource["published"]
        else:
            published = [{"datetime": web.utcnow(),
                          "timezone": "America/Los_Angeles"}]
        resource["visibility"] = resource.get("visibility", ["public"])
        resource["channel"] = resource.get("channel", ["default"])
        mentions = []
        urls = resource.pop("url", [])
        if resource_type == "card":
            slug = resource.get("nickname", resource.get("name"))[0]
            urls.insert(0, f"/pub/cards/{web.textslug(slug)}")
            resource.update(published=published)
        elif resource_type == "feed":
            slug = resource["name"][0]
            if not urls:
                urls.append(f"/{web.textslug(slug)}")
            resource.pop("channel")
        elif resource_type == "entry":
            post_type = discover_post_type(resource)
            slug = None
            if post_type == "article":
                slug = resource["name"][0]
            elif post_type == "bookmark":
                mentions.append(resource["bookmark-of"][0])
            elif post_type == "like":
                mentions.append(resource["like-of"][0])
            elif post_type == "rsvp":
                mentions.append(resource["in-reply-to"][0])
            # elif post_type == "identification":
            #     identifications = resource["identification-of"]
            #     identifications[0] = {"type": "cite",
            #                           **identifications[0]["properties"]}
            #     textslug = identifications[0]["name"]
            #     mentions.append(identifications[0]["url"])
            # elif post_type == "follow":
            #     follows = resource["follow-of"]
            #     follows[0] = {"type": "cite", **follows[0]["properties"]}
            #     textslug = follows[0]["name"]
            #     mentions.append(follows[0]["url"])
            #     tx.sub.follow(follows[0]["url"])
            path = f"/{web.timeslug(published[0]['datetime'])}"
            if slug:
                path += f"/{web.textslug(slug)}"
            urls.append(path)
            author_id = tx.db.select("resources",
                                     where="""json_extract(resources.resource,
                                                  '$.uid[0]') == ?""",
                                     vals=[tx.origin])[0]["version"]
            resource.update(author=[author_id])
        # elif resource_type == "event":
        #     slug = resource.get("nickname", resource.get("name"))[0]
        #     urls.insert(0, f"/pub/cards/{web.textslug(slug)}")
        #     # if resource["uid"] == str(web.uri(tx.host.name)):
        #     #     pass
        resource.update(published=published, url=urls)
        resource["type"] = [resource_type]
        resource_kwargs = dict(resource=resource, version=web.nbrandom(10))
        from pprint import pprint
        pprint(resource_kwargs)
        tx.db.insert("resources", **resource_kwargs)
        web.publish("/recent", ".feed[-0:-0]", resource)
        for mention in mentions:
            web.enqueue(webmention.send, f"{tx.origin}{path}", mention)
        # TODO web.publish(mention, ".responses[-1:-1]", resource)
        # for subscriber in subscribers:
        #     web.enqueue(websub.publish)
        return urls[0]

    def update(self, url, add=None, replace=None, remove=None):
        """Update a resource."""
        now = web.utcnow()
        path = f"/{url.strip('/')}"
        resource = tx.db.select("resources", where="path = ?",
                                vals=[path])[0]["resource"]
        for prop, vals in add.items():
            try:
                resource[prop].extend(vals)
            except KeyError:
                resource[prop] = vals
        resource["updated"] = now
        tx.db.update("resources", resource=resource,
                     where="path = ?", vals=[path])
        # TODO web.publish(url, f".{prop}[-0:-0]", vals)

    def delete(self, url):
        """Delete a resource."""
        pass  # TODO

    def search(self, query):
        """Return a list of resources containing `query`."""
        where = """json_extract(resources.resource,
                       '$.bookmark-of[0].url') == ?
                   OR json_extract(resources.resource,
                       '$.like-of[0].url') == ?"""
        return tx.db.select("resources", vals=[query, query], where=where)

    def read(self, url):
        """Return an entry with its metadata."""
        if not url.startswith(("http://", "https://")):
            url = f"/{url.rstrip('/')}"
        try:
            resource = tx.db.select("resources",
                                    where="""json_extract(
                                                 resources.resource,
                                                 '$.url[0]') == ?""",
                                    vals=[url])[0]
        except IndexError:
            try:
                resource = tx.db.select("resources",
                                        where="""json_extract(
                                                     resources.resource,
                                                     '$.alias[0]') == ?""",
                                        vals=[url])[0]
            except IndexError:
                web.header("Content-Type", "text/html")  # TODO FIXME XXX
                raise web.NotFound(templates.post_not_found(url))
        r = resource["resource"]
        if "entry" in r["type"]:
            r["author"] = self.get_version(r["author"][0])["resource"]
        return resource

    def get_version(self, version):
        """Return a snapshot of resource at given version."""
        return tx.db.select("resources", where="version = ?",
                            vals=[version])[0]

    def get_entry(self, path):
        """"""

    def get_card(self, nickname):
        """Return the card with given nickname."""
        resource = tx.db.select("resources", vals=[nickname],
                                where="""json_extract(resources.resource,
                                             '$.nickname[0]') == ?""")[0]
        return resource["resource"]

    def get_event(self, path):
        """"""

    def get_entries(self, limit=20, modified="DESC"):
        """Return a list of entries."""
        return tx.db.select("resources",
                            order=f"""json_extract(resources.resource,
                                          '$.published[0]') {modified}""",
                            where="""json_extract(resources.resource,
                                         '$.type[0]') == 'entry'""",
                            limit=limit)

    def get_cards(self, limit=20):
        """Return a list of alphabetical cards."""
        return tx.db.select("resources",  # order="modified DESC",
                            where="""json_extract(resources.resource,
                                         '$.type[0]') == 'card'""")

    def get_rooms(self, limit=20):
        """Return a list of alphabetical rooms."""
        return tx.db.select("resources",  # order="modified DESC",
                            where="""json_extract(resources.resource,
                                         '$.type[0]') == 'room'""")

    def get_channels(self):
        """Return a list of alphabetical channels."""
        return tx.db.select("resources",  # order="modified DESC",
                            where="""json_extract(resources.resource,
                                         '$.type[0]') == 'feed'""")

    def get_posts_by_channel(self, uid):
        """Return the card with given nickname."""
        return tx.db.select("resources", vals=[uid],
                            where="""json_extract(resources.resource,
                                         '$.channel[0]') == ?""",
                            order="""json_extract(resources.resource,
                                         '$.published[0]') DESC""")

    # def get_channels(self):
    #     """Return a list of channels."""
    #     return [r["value"] for r in
    #             tx.db.select("""resources,
    #                            json_tree(resources.resource, '$.channel')""",
    #                          what="DISTINCT value", where="type = 'text'")]

    def get_year(self, year):
        return tx.db.select("resources",
                            order="""json_extract(resources.resource,
                                         '$.published[0].datetime') ASC""",
                            where=f"""json_extract(resources.resource,
                                          '$.published[0].datetime')
                                          LIKE '{year}%'""")

    def get_media(self):
        """Return a list of media filepaths."""
        try:
            filepaths = list(pathlib.Path(tx.host.name).iterdir())
        except FileNotFoundError:
            filepaths = []
        return filepaths

    def get_filepath(self, filename):
        """Return a media file's path."""
        return pathlib.Path(tx.host.name) / filename


@server.route(r"")
class MicropubEndpoint:
    """."""

    def get(self):
        local_client = LocalClient()
        try:
            form = web.form("q")
        except web.BadRequest:
            channels = local_client.get_channels()
            entries = local_client.get_entries()
            cards = local_client.get_cards()
            events = []  # local_client.get_events()
            reviews = []  # local_client.get_reviews()
            rooms = local_client.get_rooms()
            media = local_client.get_media()
            return templates.activity(channels, entries, cards, events,
                                      reviews, rooms, media)
        syndication_endpoints = []

        def generate_channels():
            return [{"name": r["name"][0],
                     "uid": r["uid"][0]} for r in local_client.get_channels()]

        if form.q == "config":
            response = {"q": ["category", "contact", "source", "syndicate-to"],
                        "media-endpoint": f"https://{tx.host.name}/pub/media",
                        "syndicate-to": syndication_endpoints,
                        "visibility": ["public", "unlisted", "private"],
                        "channels": generate_channels()}
        elif form.q == "channel":
            response = {"channels": generate_channels()}
        elif form.q == "source":
            response = {}
            if "search" in form:
                response = {"items": [{"url": [r["resource"]["url"]]} for r in
                                      local_client.search(form.search)]}
            if "url" in form:
                response = dict(local_client.read(form.url))
        else:
            raise web.BadRequest("""unsupported query.
                                    check `q=config` for support.""")
        web.header("Content-Type", "application/json")
        return response

    def post(self):
        try:
            form = web.form("h")
        except web.BadRequest:
            resource = tx.request.body._data
        else:
            h = form.pop('h')
            properties = {k.rstrip("[]"): (v if isinstance(v, list) else [v])
                          for k, v in form.items()}
            resource = {"type": [f"h-{h}"], "properties": properties}
        client = LocalClient()
        action = resource.pop("action", None)
        if action == "update":
            url = resource.pop("url")
            client.update(url, **resource)
            return
        permalink = client.create(resource["type"][0].partition("-")[2],
                                  **resource["properties"])
        # web.header("Link", '</blat>; rel="shortlink"', add=True)
        # web.header("Link", '<https://twitter.com/angelogladding/status/'
        #                    '30493490238590234>; rel="syndication"', add=True)
        raise web.Created("post created", permalink)


@server.route(r"channels")
class Channels:
    """All channels."""

    def get(self):
        return templates.channels(LocalClient().get_channels())


@server.route(r"channels/{channel}")
class Channel:
    """A single channel."""

    def get(self):
        return templates.channel(self.channel)


@server.route(r"entries")
class Entries:
    """All entries on file."""

    def get(self):
        return templates.entries(LocalClient().get_entries(),
                                 templates.render_dict)


@server.route(r"entries/{entry}")
class Entry:
    """A single entry on file."""

    def get(self):
        try:
            resource = tx.db.select("cache", where="url = ?",
                                    vals=[f"https://{self.resource}"])[0]
        except IndexError:
            resource = tx.db.select("cache", where="url = ?",
                                    vals=[f"http://{self.resource}"])[0]
        return templates.cache_resource(resource)


@server.route(r"cards")
class Cards:
    """
    All cards on file.

    `OPTIONS`, `PROPFIND` and `REPORT` methods provide CardDAV support.

    """

    def get(self):
        return templates.cards(tx.pub.get_cards(), templates.render_dict)

    def options(self):
        """Signal capabilities to CardDAV client."""
        web.header("DAV", "1, 2, 3, access-control, addressbook")
        web.header("Allow", "OPTIONS, GET, HEAD, POST, PUT, DELETE, TRACE, "
                            "COPY, MOVE, MKCOL, PROPFIND, PROPPATCH, LOCK, "
                            "UNLOCK, REPORT, ACL")
        tx.response.naked = True
        return ""

    def propfind(self):
        """
        Return a status listing of addressbook/contacts.

        This resource is requsted twice with `Depth` headers of 0 and 1.
        0 is a request for the addressbook itself. 1 is a request for the
        addressbook itself and all contacts in the addressbook. Thus both
        the addressbook itself and each user have an etag.

        """
        # TODO refactor..
        web.header("DAV", "1, 2, 3, access-control, addressbook")

        depth = int(tx.request.headers["Depth"])
        etags = {"": tx.kv["carddav-lasttouch"]}
        if depth == 1:
            for identity in get_resources("identities"):
                etags[identity["-uuid"]] = \
                    identity.get("updated", identity["published"]).timestamp()

        props = list(tx.request.body.iterchildren())[0]
        namespaces = set()
        responses = []

        for uuid, etag in etags.items():
            ok = []
            notfound = []
            for prop in props.iterchildren():
                # supported
                if prop.tag == "{DAV:}current-user-privilege-set":
                    ok.append("""<current-user-privilege-set>
                                     <privilege>
                                         <all />
                                         <read />
                                         <write />
                                         <write-properties />
                                         <write-content />
                                     </privilege>
                                 </current-user-privilege-set>""")
                if prop.tag == "{DAV:}displayname":
                    ok.append("<displayname>carddav</displayname>")
                if prop.tag == "{DAV:}getetag":
                    ok.append(f'<getetag>"{etag}"</getetag>')
                if prop.tag == "{DAV:}owner":
                    ok.append("<owner>/</owner>")
                if prop.tag == "{DAV:}principal-URL":
                    ok.append("""<principal-URL>
                                     <href>/identities</href>
                                 </principal-URL>""")
                if prop.tag == "{DAV:}principal-collection-set":
                    ok.append("""<principal-collection-set>
                                     <href>/identities</href>
                                 </principal-collection-set>""")
                if prop.tag == "{DAV:}current-user-principal":
                    ok.append("""<current-user-principal>
                                     <href>/identities</href>
                                 </current-user-principal>""")
                if prop.tag == "{DAV:}resourcetype":
                    namespaces.add("CR")
                    if uuid:
                        ok.append("<resourcetype />")
                    else:
                        ok.append("""<resourcetype>
                                         <CR:addressbook />
                                         <collection />
                                     </resourcetype>""")
                if prop.tag == "{DAV:}supported-report-set":
                    ok.append("""<supported-report-set>
                                     <supported-report>
                                         <report>principal-property-search</report>
                                     </supported-report>
                                     <supported-report>
                                         <report>sync-collection</report>
                                     </supported-report>
                                     <supported-report>
                                         <report>expand-property</report>
                                     </supported-report>
                                     <supported-report>
                                         <report>principal-search-property-set</report>
                                     </supported-report>
                                 </supported-report-set>""")
                if (prop.tag == "{urn:ietf:params:xml:ns:carddav}"
                                "addressbook-home-set"):
                    namespaces.add("CR")
                    ok.append("""<CR:addressbook-home-set>
                                     <href>/identities</href>
                                 </CR:addressbook-home-set>""")
                if (prop.tag == "{http://calendarserver.org/ns/}"
                                "getctag"):
                    namespaces.add("CS")
                    ok.append(f'<CS:getctag>"{etag}"</CS:getctag>')

                # conditionally supported
                if (prop.tag == "{http://calendarserver.org/ns/}me-card"):
                    namespaces.add("CS")
                    if uuid:
                        notfound.append("<CS:me-card />")
                    else:
                        ok.append(f"""<CS:me-card>
                                      <href>/identities/{tx.owner["-uuid"]}.vcf</href>
                                      </CS:me-card>""")

                # not supported
                if prop.tag == "{DAV:}add-member":
                    notfound.append("<add-member />")
                if prop.tag == "{DAV:}quota-available-bytes":
                    notfound.append("<quota-available-bytes />")
                if prop.tag == "{DAV:}quota-used-bytes":
                    notfound.append("<quota-used-bytes />")
                if prop.tag == "{DAV:}resource-id":
                    notfound.append("<resource-id />")
                if prop.tag == "{DAV:}sync-token":
                    notfound.append("<sync-token />")
                if (prop.tag == "{urn:ietf:params:xml:ns:carddav}"
                                "directory-gateway"):
                    namespaces.add("CR")
                    notfound.append("<CR:directory-gateway />")
                if (prop.tag == "{urn:ietf:params:xml:ns:carddav}"
                                "max-image-size"):
                    namespaces.add("CR")
                    notfound.append("<CR:max-image-size />")
                if (prop.tag == "{urn:ietf:params:xml:ns:carddav}"
                                "max-resource-size"):
                    namespaces.add("CR")
                    notfound.append("<CR:max-resource-size />")
                if (prop.tag == "{http://calendarserver.org/ns/}"
                                "email-address-set"):
                    namespaces.add("CS")
                    notfound.append("<CS:email-address-set />")
                if (prop.tag == "{http://calendarserver.org/ns/}"
                                "push-transports"):
                    namespaces.add("CS")
                    notfound.append("<CS:push-transports />")
                if (prop.tag == "{http://calendarserver.org/ns/}"
                                "pushkey"):
                    namespaces.add("CS")
                    notfound.append("<CS:pushkey />")
                if (prop.tag == "{http://me.com/_namespace/}"
                                "bulk-requests"):
                    namespaces.add("ME")
                    notfound.append("<ME:bulk-requests />")
            href = "/identities"
            if uuid:
                href += f"/{uuid}.vcf"
            responses.append((href, ok, notfound))
        tx.response.naked = True
        raise web.MultiStatus(view.carddav(namespaces, responses))

    def report(self):
        """Return a full listing for each requested identity."""
        etags = {}
        for identity in get_resources("identities"):
            etags[identity["-uuid"]] = \
                identity.get("updated", identity["published"]).timestamp()
        children = list(tx.request.body.iterchildren())
        # XXX props = children[0]  # TODO soft-code prop responses
        responses = []
        for href in children[1:]:
            uuid = href.text.rpartition("/")[2].partition(".")[0]
            ok = [f'<getetag>"{etags[uuid]}"</getetag>',
                  f"<CR:address-data>{generate_vcard(uuid)}</CR:address-data>"]
            notfound = []
            responses.append((href.text, ok, notfound))
        tx.response.naked = True
        raise web.MultiStatus(view.carddav(set(["CR"]), responses))


@server.route(r"cards/{nickname}")
class Card:
    """A single card on file."""

    def get(self):
        # try:
        #     resource = tx.db.select("cache", where="url = ?",
        #                             vals=[f"https://{self.resource}"])[0]
        # except IndexError:
        #     resource = tx.db.select("cache", where="url = ?",
        #                             vals=[f"http://{self.resource}"])[0]
        return templates.card(tx.pub.get_card(self.nickname))


@server.route(r"cards/{nickname}.vcf")
class VCard:
    """
    A single card on file, represented as a vCard.
    
    `PUT` and `DELETE` methods provide CardDAV support.

    """

    def get(self):
        web.header("Content-Type", "text/vcard")
        return generate_vcard(self.nickname)

    def put(self):
        """
        add or update a identity

        """
        # TODO only add if "if-none-match" is found and identity isn't
        try:
            print("if-none-match", tx.request.headers.if_none_match)
        except AttributeError:
            pass
        else:
            try:
                identities.get_identity_by_uuid(self.card_id)
            except ResourceNotFound:
                pass
            else:
                raise web.Conflict("identity already exists")

        # TODO only update if "if-match" matches etag on hand
        try:
            request_etag = str(tx.request.headers.if_match).strip('"')
            print("if-match", request_etag)
        except AttributeError:
            pass
        else:
            identity = identities.get_identity_by_uuid(self.card_id)
            current_etag = \
                identity.get("updated", identity["published"]).timestamp()
            print("current etag", current_etag)
            if request_etag != current_etag:
                raise web.Conflict("previous edit already exists")

        # TODO non-standard type-params (url) not handled by vobject

        card = vobject.readOne(tx.request.body.decode("utf-8"))

        name = card.fn.value.strip()

        extended = {}
        n = card.n.value

        def explode(key):
            item = getattr(n, key)
            if isinstance(item, list):
                extended[key] = ";".join(item)
            else:
                extended[key] = [item]
        explode("prefix")
        explode("given")
        explode("additional")
        explode("family")
        explode("suffix")

        # TODO identity_type = "identity"
        basic = {"name": name, "uuid": self.card_id}

        # TODO organizations = [o.value[0]
        # TODO                  for o in card.contents.get("org", [])]
        # TODO for organization in organizations:
        # TODO     if organization == name:
        # TODO         identity_type = "organization"

        # TODO telephones = []
        # TODO for tel in card.contents.get("tel", []):
        # TODO     telephones.append((tel.value, tel.params["TYPE"]))
        # TODO websites = []
        # TODO for url in card.contents.get("url", []):
        # TODO     type = url.params.get("TYPE", [])
        # TODO     for label in card.contents.get("x-ablabel"):
        # TODO         if label.group == url.group:
        # TODO             type.append(label.value)
        # TODO     print(url.value, type)
        # TODO     print()
        # TODO     websites.append((url.value, type))

        # photo = card.contents.get("photo")[0]
        # print()
        # print(photo)
        # print()
        # print(photo.group)
        # print(photo.params.get("ENCODING"))
        # print(photo.params.get("X-ABCROP-RECTANGLE"))
        # print(photo.params.get("TYPE", []))
        # print(len(photo.value))
        # print()
        # filepath = tempfile.mkstemp()[1]
        # with open(filepath, "wb") as fp:
        #     fp.write(photo.value)
        # photo_id = canopy.branches["images"].photos.upload(filepath)
        # extended["photos"] = [photo_id]

        try:
            details = identities.get_identity_by_uuid(self.card_id)
        except ResourceNotFound:
            print("NEW identity!")
            print(basic)
            print(extended)
            quick_draft("identity", basic,
                        publish="Identity imported from iPhone.")
            # XXX details = create_identity(access="private", uid=self.card_id,
            # XXX                          **basic)
            # XXX details = update_identity(identifier=details["identifier"],
            # XXX                  telephones=telephones, websites=websites,
            # XXX                  **extended)
            print("CREATED")
        else:
            print("EXISTING identity!")
            print(details)
            print("UPDATED")
        # XXX     basic.update(extended)
        # XXX     details = update_identity(identifier=details["identifier"],
        # XXX                      telephones=telephones, websites=websites,
        # XXX                      **basic)
        identity = identities.get_identity_by_uuid(self.card_id)
        etag = identity.get("updated", identity["published"]).timestamp()
        web.header("ETag", f'"{etag}"')
        tx.response.naked = True
        raise web.Created("created identity",
                          f"/identities/{self.card_id}.vcf")

    def delete(self):
        """
        delete a identity

        This method provides CardDAV support.

        """
        # delete_resource(...)
        tx.response.naked = True
        return f"""<?xml version="1.0"?>
                   <multistatus xmlns="DAV:">
                     <response>
                       <href>/identities/{self.card_id}.vcf</href>
                       <status>HTTP/1.1 200 OK</status>
                     </response>
                   </multistatus>"""


@server.route(r"rooms")
class Rooms:
    """All rooms on file."""

    def get(self):
        return templates.rooms(tx.pub.get_rooms(), templates.render_dict)


@server.route(r"syndication")
class Syndication:
    """."""

    def get(self):
        return templates.syndication()

    def post(self):
        destinations = web.form()
        if "twitter_username" in destinations:
            un = destinations.twitter_username
            # TODO pw = destinations.twitter_password
            # TODO sign in
            user_photo = ""  # TODO doc.qS(f"a[href=/{un}/photo] img").src
            destination = {"uid": f"//twitter.com/{un}",
                           "name": f"{un} on Twitter",
                           "service": {"name": "Twitter",
                                       "url": "//twitter.com",
                                       "photo": "//abs.twimg.com/favicons/"
                                                "twitter.ico"},
                           "user": {"name": un, "url": f"//twitter.com/{un}",
                                    "photo": user_photo}}
            tx.db.insert("syndication", destination=destination)
        if "github_username" in destinations:
            un = destinations.github_username
            # TODO token = destinations.github_token
            # TODO check the token
            user_photo = ""  # TODO doc.qS("img.avatar-user.width-full").src
            destination = {"uid": f"//github.com/{un}",
                           "name": f"{un} on GitHub",
                           "service": {"name": "GitHub",
                                       "url": "//github.com",
                                       "photo": "//github.githubassets.com/"
                                                "favicons/favicon.png"},
                           "user": {"name": un, "url": f"//github.com/{un}",
                                    "photo": user_photo}}
            tx.db.insert("syndication", destination=destination)


@server.route(r"media")
class MediaEndpoint:
    """."""

    def get(self):
        return templates.media(LocalClient().get_media())

    def post(self):
        media_dir = pathlib.Path(tx.host.name)
        media_dir.mkdir(exist_ok=True, parents=True)
        while True:
            mid = web.nbrandom(4)
            filename = media_dir / mid
            if not filename.exists():
                filename = web.form("file").file.save(filename)
                break
        if str(filename).endswith(".heic"):
            sh.convert(filename, "-set", "filename:base", "%[basename]",
                       f"{media_dir}/%[filename:base].jpg")
        sha256 = str(sh.sha256sum(filename)).split()[0]
        try:
            tx.db.insert("media", mid=mid, sha256=sha256,
                         size=filename.stat().st_size)
        except tx.db.IntegrityError:
            mid = tx.db.select("media", where="sha256 = ?",
                               vals=[sha256])[0]["mid"]
            filename.unlink()
        path = f"/pub/media/{mid}"
        raise web.Created(f"File can be found at <a href={path}>{path}</a>",
                          location=path)


@server.route(r"media/{filename}")
class MediaFile:
    """."""

    def get(self):
        content_types = {(".jpg", ".jpeg"): "image/jpg",
                         ".heic": "image/heic",
                         ".png": "image/png",
                         ".mp3": "audio/mpeg",
                         ".mov": "video/quicktime",
                         ".mp4": "video/mp4"}
        for suffix, content_type in content_types.items():
            if self.filename.endswith(suffix):
                web.header("Content-Type", content_type)
                break
        relative_path = f"{tx.host.name}/{self.filename}"
        if tx.host.server[0] == "gunicorn":
            with open(relative_path, "rb") as fp:
                return fp.read()
        else:  # assumes Nginx context
            web.header("X-Accel-Redirect", f"/X/{relative_path}")

    def delete(self):
        filepath = LocalClient().get_filepath(self.filename)
        tx.db.delete("media", where="mid = ?", vals=[filepath.stem])
        filepath.unlink()
        return "deleted"



"""
Render formatted plaintext (Markdown) as HTML.

Plaintext to HTML that strives to use concise, human-readable textual anno-
tations to build rich documents of various size and purpose including sta-
tus updates, scholarly works, screenwriting, literate programming, etc.

The guiding philosophy of both the syntax and features is to allow the
user as much freedom as possible to convey a message while automatically
handling the distracting, repetitive details commonly involved with pub-
lishing for the web.

    >>> str(render("foo bar"))
    '<p>foo bar</p>'

>   The idea was to make writing simple web pages ... as easy as writing
>   an email, by allowing you to use much the same syntax and converting
>   it automatically into HTML ... [and] back into Markdown.

--- Aaron Swartz, [Markdown][1] -- March 19, 2004

[1]: http://www.aaronsw.com/weblog/001189

"""

# TODO lettered-list
# TODO talk about typesetting
# TODO bibtex??
# TODO mathtex??
# TODO font-families
# TODO microformats
# TODO code[, var, kbd, ..](`), sub(_), sup(^), em(*), strong(**)
# TODO a, footnote[^abba] (w/ \u21A9), abbr, cite[@Angelo], q (w/ @cite)
# TODO table
# TODO img, figure (w/ figcaption), audio, video
# TODO smart quotes, dashes, ellipses
# TODO widont
# TODO syntax highlight
# TODO table of contents, reheader (# > h2), index, bibliography
# TODO papaya pilcrows
# TODO emoticons
# TODO l10n (charset)
# TODO math
# TODO timestamp
# TODO [tl;dr] as Abstract
# TODO formulae to formul\u00E6
# TODO slidy/S5

import zlib  # TODO more useful hash?

import lxml.html
import lxml.html.builder as E
try:
    import regex as re
except ImportError:
    import re

from . import block
from . import inline

__all__ = ["render"]


_LT = "%!lt!%"
_GT = "%!gt!%"
_LTS = "%!lts!%"
_GTS = "%!gts!%"
_AMP = "%!amp!%"

reference_pattern = re.compile(r"\s*\[([\w\s-]+)\]:\s+(.{2,})")
abbreviation_pattern = re.compile(r"\s*\*\[([\w\s-]+)\]:\s+(.{2,})")


def obfuscate_references(mahna):
    """TODO."""
    references = {}
    for chunk in re.split(r"\n{2,}", mahna):
        reference_match = reference_pattern.match(chunk)
        if reference_match:
            for id, uri in reference_pattern.findall(chunk):
                references.update({id + "---" + str(zlib.adler32(uri)): uri})
            continue
    for id in references.keys():
        before = "[{}]".format(id.partition("---")[0])
        after = "[{}]".format(id)
        mahna = mahna.replace(before, after)
    return mahna


def preprocess_lga(text):
    text = text.replace("<", _LT)
    text = text.replace(">", _GT)
    text = text.replace("&lt;", _LTS)
    text = text.replace("&gt;", _GTS)
    text = text.replace("&", _AMP)
    return text


def postprocess_lga(text):
    text = str(text)
    text = text.replace(_LT, "<")
    text = text.replace(_GT, ">")
    text = text.replace(_AMP, "&")
    return text


def preprocess_script_style_pre_textarea(mahna):
    def esc(m):
        return m.group(0).replace("\n", "%!nl!%")
    mahna = re.sub(r"(?s)<script.*?>(.*?)</script>", esc, mahna)
    mahna = re.sub(r"(?s)<style.*?>(.*?)</style>", esc, mahna)
    mahna = re.sub(r"(?s)<pre.*?>(.*?)</pre>", esc, mahna)
    mahna = re.sub(r"(?s)<textarea.*?>(.*?)</textarea>", esc, mahna)
    return mahna


def preprocess_inline_code(mahna):
    # TODO !!! escape from inside <pre></pre>
    return mahna
    # TODO handle "an inline code block, `\`<div>\``, preprocessor"

    def escape_match_handler(m):
        return "<code>{0}</code>".format(preprocess_lga(m.group(0).strip("`")))
    return re.sub(r"`(.*?)`", escape_match_handler, mahna)


def preprocess_hyphens(mahna):
    new_lines = []
    queue = []
    for line in mahna.splitlines():
        if line != "-" and not line.endswith("--") and line.endswith("-"):
            queue.append(line[:-1])
            continue
        queued = ""
        if len(queue) > 0:
            queued += queue[0]
            line = line.lstrip(" >")
        if len(queue) > 1:
            queued += "".join(line.strip(" >") for line in queue[1:])
        new_lines.append(queued + line)
        queue = []
    return "\n".join(new_lines)


def postprocess_script_style_pre_textarea_code(html):
    html = html.replace("%!nl!%", "\n")
    html = html.replace("%!lts!%", "&lt;")
    html = html.replace("%!gts!%", "&gt;")
    return html


class Document:
    """A Markdown parser."""

    def __init__(self, text, context=None):
        """Return a  after parsing the Markdown text."""
        self.abbrs = {}
        self.refs = {}
        self.toc = []
        self.tags = []
        self.at_mentions = []
        self.mentioned_urls = []
        self.context = context
        self.process(text)

    def process(self, mahna):  # , options=None):
        # document = render(mahna)  # TODO prerender?
        # document = "<meta charset=utf-8>\n" + mahna
        mahna = "\n\n" + str(mahna).strip()
        mahna = preprocess_script_style_pre_textarea(mahna)
        mahna = preprocess_inline_code(mahna)
        mahna = preprocess_hyphens(mahna)

        body = E.DIV()
        self.process_blocks(body, mahna)
        output = lxml.html.tostring(body, encoding="utf-8").decode("utf-8")

        output = output.replace("<p>" + _LT, "<")
        output = output.replace(_GT + "</p>", ">")
        output = postprocess_lga(output)

        dom = lxml.html.fromstring(output)
        dom = E.BODY(dom)
        dom[0].drop_tag()

        for handler_name in inline.__all__:
            handler = getattr(inline, handler_name)()
            self.proc_inlines(handler, dom)
            output = lxml.html.tostring(dom, encoding="utf-8").decode("utf-8")
            output = postprocess_lga(output)
            dom = lxml.html.fromstring(output)

        # self.process_inlines(0, dom)
        # output = lxml.html.tostring(dom, encoding="utf-8").decode("utf-8")
        # output = postprocess_lga(output)
        # dom = lxml.html.fromstring(output)

        dom = E.BODY(dom)

        # TODO repeat setup and teardown for each inline handler as hacked
        #      below for hyphenation -- isolate link creation as root cause?

        # self.process_inlines(1, dom)
        # output = lxml.html.tostring(dom, encoding="utf-8").decode("utf-8")
        # output = postprocess_lga(output)
        # dom = lxml.html.fromstring(output)

        self.process_abbreviations(dom)

        output = lxml.html.tostring(dom, pretty_print=True,
                                    encoding="utf-8").decode("utf-8")
        output = lxml.html.tostring(dom, pretty_print=True).decode("utf-8")

        output = postprocess_lga(output)
        output = postprocess_script_style_pre_textarea_code(output)

        # XXX here be dragons -- skips jumps in hierarchy (e.g. h5 follows h3)
        top_heading_level = 1

        def tocify(_toc, curr):
            section = E.OL()
            i = 0
            for level, text, id in _toc:
                i += 1
                if level == top_heading_level:
                    continue
                if level < curr:
                    return section
                if level == curr:
                    section.append(E.LI(E.A(text, href="#" + id),
                                        tocify(_toc[i:], curr + 1)))
            return section

        toc = tocify(self.toc, top_heading_level + 1)
        toc.set("id", "toc")  # toc["id"] = "toc"
        for ol in toc.cssselect("ol"):    # XXX two words, rhymes with bucket
            if ol.getparent() is not None and not len(ol):
                ol.drop_tag()
        output = output.replace("<p>[Contents] </p>",
                                str(lxml.html.tostring(toc)))

        # if options and options.document:
        #     try:
        #         title = document.title
        #     except AttributeError:
        #         title = body.cssselect("h1")[0].text_content()
        #     self.stylesheet = "<style>{0}</style>".format(self.stylesheet)
        #     try:
        #         style = self.stylesheet + document.stylesheet
        #     except AttributeError:
        #         style = self.stylesheet
        #     output = html.template(options.language, title, style, output)
        # XXX sometimes `lxml` wraps with a <div> .. sometimes it doesn't
        # XXX probably something to do with fromstring/tostring    fffuuuuuu
        # elif output.startswith("<div>"):
        output = output.strip()
        if output.startswith("<body>"):
            output = output[6:-7]
        output = output.strip()
        if output.startswith("<div>"):
            output = output[5:-6]
        output = output.strip()
        if output.startswith("<p></p>"):
            output = output[8:]
        # XXX if inline:
        # XXX     html = html[html.find(">")+1:html.rfind("<")]
        self.html = output

    def process_blocks(self, parent, mahna):
        for chunk in re.split(r"\n{2,}", mahna):
            reference_match = reference_pattern.match(chunk)
            if reference_match:
                self.refs.update(reference_pattern.findall(chunk))
                continue
            abbreviation_match = abbreviation_pattern.match(chunk)
            if abbreviation_match:
                self.abbrs.update(abbreviation_pattern.findall(chunk))
                continue
            for name in block.__all__:
                handler = getattr(block, name)(self)
                match = re.match(handler.pattern, chunk)
                if match:
                    # XXX try:
                    # XXX     encoded_chunk = chunk.decode("utf-8")
                    # XXX except UnicodeEncodeError:
                    # XXX     encoded_chunk = chunk
                    encoded_chunk = chunk

                    # try:
                    element = handler.parse(encoded_chunk)
                    if element is not None:
                        parent.append(element)
                    # except ValueError:
                    #     chunk = "".join(c if ord(c) < 128 else
                    #                     c+"!?" for c in chunk)
                    #     print("Found non Unicode or ASCII char in chunk:",
                    #                 chunk, sep="\n\n", file=sys.stderr)
                    #     sys.exit(1)
                    break

    def proc_inlines(self, handler, parent):
        for child in parent:
            if child.tag in ("script", "style", "pre", "code", "textarea"):
                continue
            handler.parser = self
            self.child = child
            if child.text:
                child.text = re.sub(handler.pattern, handler.parse, child.text)
            if child.tail:
                child.tail = re.sub(handler.pattern, handler.parse, child.tail)
            if child.text:
                child.text = preprocess_lga(child.text)
            if child.tail:
                child.tail = preprocess_lga(child.tail)
            self.proc_inlines(handler, child)

    # def process_inlines(self, stage, parent):
    #     for child in parent:
    #         if child.tag in ("script", "style", "pre", "code", "textarea"):
    #             continue
    #         handler_names = inline.__all__ if stage == 0 else ["Hyphenate"]
    #         for handler_name in handler_names:
    #             handler = getattr(inline, handler_name)()
    #             handler.parser = self
    #             # XXX self.child = child
    #             # XXX self.parent = parent

    #             if child.text:
    #                 while True:
    #                     found = re.split(f"({handler.split_pattern})",
    #                                      child.text)
    #                     print()
    #                     print(handler_name, child.text)
    #                     # print(handler.pattern)
    #                     # print(child.text)
    #                     print(found)
    #                     if len(found) < 3:
    #                         break
    #                     before = "".join(found[:-2])
    #                     matched_text, after = found[-2:]
    #                     child.text = before
    #                     match = re.match(handler.pattern, matched_text)
    #                     replacement = handler.parse(match)
    #                     # if isinstance(replacement, lxml.html.HtmlElement):
    #                     replacement.tail = after
    #                     child.insert(0, replacement)
    #                     # print(lxml.html.tostring(replacement))
    #                     # else:
    #                     #     child.text += replacement + after
    #                     # print(child.text)
    #                     # print(lxml.html.tostring(child))

    #             # if child.text:
    #             #     child.text = re.sub(handler.pattern, handler.parse,
    #             #                         child.text)

    #             # if child.tail:
    #             #     child.tail = re.sub(handler.pattern, handler.parse,
    #             #                         child.tail)
    #         if child.text:
    #             child.text = preprocess_lga(child.text)
    #         if child.tail:
    #             child.tail = preprocess_lga(child.tail)
    #         self.process_inlines(stage, child)

    def process_abbreviations(self, parent):
        for child in parent:
            if child.tag in ("script", "style", "pre", "code", "textarea"):
                continue
            for abbr, full in self.abbrs.items():
                html = '<abbr title="{}">{}</abbr>'.format(full, abbr)
                if child.text:
                    child.text = preprocess_lga(child.text.replace(abbr, html))
                if child.tail:
                    child.tail = preprocess_lga(child.tail.replace(abbr, html))
            self.process_abbreviations(child)

    def __str__(self):
        return self.html


render = Document



"""
tools for NewBase60, NewBase64 and NewDateTime numbering systems

NewBase60
---------

NewBase60 is a base 60 (sexagesimal) numbering system that uses only
ASCII numbers and letters that are print, prose and code safe. Lowercase
l and uppercase I are aliased to the number 1. Capital O is aliased to
the number 0.

    0123456789ABCDEFGHJKLMNPQRSTUVWXYZ_abcdefghijkmnopqrstuvwxyz

NewBase60 encoded numbers sort properly in the typical case.

NewBase64
---------

NewBase64 is a base 64 numbering system that extends NewBase60 for use
in shortly encoding square grids. The four characters added are -, +, *
and $.

        0 1 2 3 4 5 6 7
       ----------------
     0| 0 1 2 3 4 5 6 7
     8| 8 9 A B C D E F
    16| G H J K L M N P
    24| Q R S T U V W X
    32| Y Z _ a b c d e
    40| f g h i j k m n
    48| o p q r s t u v
    56| w x y z - + * $

latitude: -90 to 90 (pole to pole)
longitude: -180 to 180 (meridian back to meridian)

Earth is split into 32 quadrants:

    0 1 2 3  4 5 6 7
    8 9 A B  C D E F
    G H J K  L M N P
    Q R S T  U V W X

Each quadrant can be further scoped by 1/64ths. e.g. the 0th quadrant:

    0 1 2 3 4 5 6 7
    8 9 A B C D E F
    G H J K L M N P
    Q R S T U V W X
    Y Z _ a b c d e
    f g h i j k m n
    o p q r s t u v
    w x y z - + * $

The initial quadrant is divided into 64 subquadrants with dimensions
5.625 degrees square. Second and third subquadrants are 0.703125 and
0.087890625 degrees square respectively.

1 degree latitude = between 110.567 (at equator) and 111.699 km (at poles)
1 degree longitude = between 111.321 (at equator) and 0 (at poles)

    precision  scope
            1  4.950 Mm

            2  618.75 km
            3  77.34375 km
            4  9.66796875 km
            5  1.20849609375 km

            6  151.06201171 m
            7  18.8827514648 m
            8  2.36034393311 m

            9  295.04299163 mm
           10  36.88037395 mm
           11  4.61004674 mm

           12  576.25584 μm
           13  72.0319803804 μm
           14  9.00399754755 μm
           15  1.12549969344 μm

NewDateTime
-----------

Four sundays (0th day of the week) mark solar equinoxes and solstices.
Monday through Saturday are days 1-6. Friday and Saturday are weekends
resulting in a "4 day work week".

5 days * 52 weeks = 260 work days in existing calendar
5 days * 50 weeks = 250 work days accounting for two weeks vacation
4 days * 60 weeks = 240 work days in NewDateTime

Four 90 day, 15 week quarters. Each quarter is split into three months.

52 days * 4 quarters = 184 work days accounting for mid-quarter vacation

        Apr  May  Jun  Jul  Aug  Sep  Oct  Nov  Dec  Jan  Feb  Mar

       0123456789ABCDEFGHJKLMNPQRSTUVWXYZ_abcdefghijkmnopqrstuvwxyz
    Su @--------------@--------------@--------------@-------------- 0
    Mo X              X              X              X               1
    Tu X              X              X              X               2
    We X    X    X    X    X    X    X    X    X    X    X    X     3
    Th X    X    X    X    X    X    X    X    X    X    X    X     4
    Fr XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX 5
    Sa XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX 6

    @ - sunday
    X - non-work

      0  year 1

    000  new epoch

    000  Sunday

    000  northern tilt - new first sunday (old Mar 20, equinox)
    0F0  northern sun - new second sunday (old Jun 21, northern solstice)
    0W0  southern tilt - new third sunday (old Sep 23, equinox)
    0k0  southern sun - new fourth sunday (old Dec 21, southern solstice)

    001  Monday
    002  Tuesday
    003  Wednesday
    004  Thursday
    005  Friday
    006  Saturday

    0z6  last day of first year

    100  second new new year

      1  year 2
     1R  26th week of the 2nd year
    1R0  Monday of the 26th week of the 2nd year

Based upon @[Tantek Çelik][1]'s [NewBase60][2], [NewBase64][3] and
[NewCalendar][4] specifications.

[1]: https://tantek.com
[2]: https://tantek.pbworks.com/NewBase60
[3]: https://tantek.pbworks.com/NewBase64
[4]: https://tantek.pbworks.com/NewCalendar

"""

import Crypto.Random
import pendulum


nb60 = "0123456789ABCDEFGHJKLMNPQRSTUVWXYZ_abcdefghijkmnopqrstuvwxyz"
nb60_re = "[0-9A-HJ-NP-Z_a-km-z]"
_nb60map = dict(zip(nb60, range(len(nb60))))
_nb60map["l"] = 1  # typo l to 1
_nb60map["I"] = 1  # typo I to 1
_nb60map["O"] = 0  # typo O to 0


def nbencode(number, min_length=1):
    """
    return the NewBase60 string equivalent of decimal integer `i`

    `min_length`: final size of string, zero padding on the left if needed

        >>> nbencode(2011)
        'ZX'
        >>> nbencode(2011, min_length=5)
        '000ZX'
        >>> nbencode(-2011)
        '-ZX'

    """
    string = ""
    try:
        number = int(number)
    except ValueError:
        raise ValueError("NewBase60 encoding requires a decimal integer")
    sign = ""
    if number < 0:
        sign = "-"
        number = abs(number)
    while number > 0:
        number, i = divmod(number, 60)
        string = nb60[i] + string
    length = len(string)
    min_length = int(min_length)
    if length < min_length:
        string = "0" * (min_length - length) + string
    return sign + string


def nbdecode(string):
    """
    return the decimal integer equivalent of NewBase60 string `s`

        >>> nbdecode('ZX')
        2011
        >>> nbdecode('000ZX')
        2011
        >>> nbdecode('-ZX')
        -2011

    """
    string = str(string)
    sign = 1
    if string.startswith("-"):
        sign = -1
        string = string[1:]
    number = 0
    for character in string:
        number = number * 60 + _nb60map.get(character, 0)
    return number * sign


def nbrandom(length):
    """
    return a random NewBase60 string `length` digits long

        >>> nbrandom(3)  # doctest: +SKIP
        'Do_'
        >>> nbrandom(6)  # doctest: +SKIP
        'j9cQ4m'

    """
    start = nbdecode("1" + "0" * (length - 1))
    end = nbdecode("z" * length)
    return nbencode(Crypto.Random.random.randint(start, end))


def nbrange(start, stop=None, step=1):
    """
    return a generator that yields NewBase60 strings like the builtin `range`

    Arguments will be coerced to their NewBase60 string representations if
    not provided as such.

        >>> list(nbrange(5))
        ['0', '1', '2', '3', '4']
        >>> list(nbrange('ex', 'f3'))
        ['ex', 'ey', 'ez', 'f0', 'f1', 'f2']

    """
    if stop is None:
        stop = start
        start = "0"
    for i in range(nbdecode(start), nbdecode(stop), step):
        yield nbencode(i)


def ncencode(dt):
    """

    """
    ddd = int(dt.format("DDD"))
    bimnumber = int((ddd - 1) / 61) + 1
    bim = hex(bimnumber + 9)[2].upper()
    dayinbim = int((ddd - 1) % 61) + 1
    seconds = round((dt.hour * 3600) + (dt.minute * 60) + dt.second) / 24
    return nbencode(dt.year), bim, nbencode(dayinbim), nbencode(seconds)


def ncdecode(year, bim, dayinbim, seconds):
    """

    """
    bimnumber = int("0x" + bim, 16) - 9
    ddd = 61 * (bimnumber - 1) + dayinbim
    return pendulum.parse("{}-{:03}".format(year, ddd))



"""A Microformats parser and utilities."""

import datetime
import json

import mf2py
import mf2util as util
from mf2util import interpret_feed, interpret_comment, interpret_event
import requests

from understory import uri

__all__ = ["parse", "util", "interpret_feed", "interpret_entry",
           "interpret_event", "interpret_comment"]


stable = {"adr": """p-street-address p-extended-address
                    p-post-office-box p-locality p-region
                    p-postal-code p-country-name p-label p/u-geo
                    p-latitude p-longitude p-altitude""",
          "card": """p-name p-honorific-prefix p-given-name
                     p-additional-name p-family-name p-sort-string
                     p-honorific-suffix p-nickname u-email u-logo
                     u-photo u-url u-uid p-category p/h-adr
                     p-post-office-box p-extended-address
                     p-street-address p-locality p-region
                     p-postal-code p-country-name p-label
                     p/u/h-geo p-latitude p-longitude p-altitude
                     p-tel p-note dt-bday u-key p-org p-job-title
                     p-role u-impp p-sex p-gender-identity
                     dt-anniversary""",
          "entry": """p-name p-summary e-content dt-published
                      dt-updated p-author p-category u-url u-uid
                      p-location u-syndication u-in-reply-to
                      p-rsvp u-like-of u-repost-of""",
          "event": """p-name p-summary dt-start dt-end dt-duration
                      e-content u-url p-category
                      p-location(card/adr/geo) [p-attendee]""",
          "feed": """p-name p-author(card) u-url u-photo""",
          "geo": """p-latitude p-longitude p-altitude""",
          "item": """p-name u-url u-photo""",
          "product": """p-name u-photo p-brand(card) p-category
                        e-content u-url u-identifier p-review(review)
                        p-price""",
          "recipe": """p-name p-ingredient p-yield e-instructions
                       dt-duration u-photo p-summary p-author(card)
                       dt-published p-nutrition p-category""",
          "resume": """p-name p-summary p-contact
                       p-education(event+card)
                       p-experience(event+card)
                       p-skill p-affiliation""",
          "review": """p-name
                       p-item(card/event/adr/geo/product/item)
                       p-author(card) dt-published p-rating p-best
                       p-worst e-content p-category u-url""",
          "review-aggregate": """p-item(card/event/adr/geo/product/item)
                                 p-average p-best p-worst p-count
                                 p-votes p-name"""}
draft = {"app": """p-name u-url u-logo u-photo"""}

for _format, _properties in stable.items():
    stable[_format] = _properties.split()
for _format, _properties in draft.items():
    draft[_format] = _properties.split()


def parse(doc=None, url=None):
    """
    Return a dictionary containing the mf2json of the HTML document `doc`.

    You may provide a document, a URL or both. When both are provided
    the URL is used as the document's base href.

    """
    url = str(url)
    if doc is None:
        try:
            response = requests.get(url)
        except requests.exceptions.SSLError:
            response = requests.get(url, verify=False)
        doc = response.text
    data = mf2py.parse(doc, url)
    data.pop("debug")
    return data


class JSONEncoder(json.JSONEncoder):

    def default(self, obj):
        if isinstance(obj, uri.URI):
            return str(obj)
        if isinstance(obj, (datetime.date, datetime.datetime)):
            return obj.isoformat()
        return json.JSONEncoder.default(self, obj)


def representative_hcard(parsed, source_url):
    """
    Find the representative hcard for a URL.

    http://microformats.org/wiki/representative-h-card-parsing

    :param dict parsed: an mf2 parsed dict
    :param str source_url: the source of the parsed document.
    :return: the representative h-card if one is found

    """
    # FIXME NOTE mf2py needs plain mf2json
    parsed = json.loads(JSONEncoder().encode(parsed))  # FIXME

    source_url = uri.parse(source_url).minimized

    def get_normal(hcard, prop):
        for _url in hcard["properties"].get(prop, []):
            try:
                yield uri.parse(_url).minimized
            except ValueError:
                pass

    hcards = [h for h in util.find_all_entries(parsed, ['h-card'],
                                               include_properties=True)
              if (h["properties"].get("name", [""])[0] or
                  h["properties"].get("nickname", [""])[0])]

    # uid and url both match source_url
    for hcard in hcards:
        if (source_url in get_normal(hcard, "uid")
                and source_url in get_normal(hcard, "url")):
            return hcard["properties"]

    # url that is also a rel=me
    for hcard in hcards:
        rel_mes = set()
        for rel_me in parsed.get("rels", {}).get("me", []):
            try:
                rel_me = uri.parse(rel_me)
            except ValueError:
                continue
            if isinstance(rel_me, (uri.HTTPURI, uri.HTTPSURI)):
                rel_mes.add(rel_me.minimized)
        if any(url in rel_mes for url in get_normal(hcard, "url")):
            return hcard["properties"]

    # single hcard with matching url
    found = []
    count = 0
    for hcard in hcards:
        # if source_url in hcard['properties'].get('url', []):
        for card_url in get_normal(hcard, "url"):
            if card_url.rstrip("/") == source_url:
                found.append(hcard)
                count += 1
    if count:
        return found[0]["properties"]


util.representative_hcard = representative_hcard


def interpret_entry(
        parsed, source_url, base_href=None, hentry=None,
        use_rel_syndication=True, want_json=False, fetch_mf2_func=None):
    """
    Given a document containing an h-entry, return a dictionary.

        {'type': 'entry',
         'url': permalink of the document (may be different than source_url),
         'published': datetime or date,
         'updated': datetime or date,
         'name': title of the entry,
         'content': body of entry (contains HTML),
         'author': {
          'name': author name,
          'url': author url,
          'photo': author photo
         },
         'syndication': [
           'syndication url',
           ...
         ],
         'in-reply-to': [...],
         'like-of': [...],
         'repost-of': [...]}

    :param dict parsed: the result of parsing a document containing mf2 markup
    :param str source_url: the URL of the parsed document, used by the
      authorship algorithm
    :param str base_href: (optional) the href value of the base tag
    :param dict hentry: (optional) the item in the above document
      representing the h-entry. if provided, we can avoid a redundant
      call to find_first_entry
    :param boolean use_rel_syndication: (optional, default True) Whether
      to include rel=syndication in the list of syndication sources. Sometimes
      useful to set this to False when parsing h-feeds that erroneously include
      rel=syndication on each entry.
    :param boolean want_json: (optional, default False) if true, the result
      will be pure json with datetimes as strings instead of python objects
    :param callable fetch_mf2_func: (optional) function to fetch mf2 parsed
      output for a given URL.
    :return: a dict with some or all of the described properties

    """
    # find the h-entry if it wasn't provided
    if not hentry:
        hentry = util.find_first_entry(parsed, ['h-entry'])
        if not hentry:
            return {}

    result = _interpret_common_properties(
        parsed, source_url, base_href, hentry, use_rel_syndication, want_json,
        fetch_mf2_func)
    if 'h-cite' in hentry.get('type', []):
        result['type'] = 'cite'
    else:
        result['type'] = 'entry'

    # NOTE patch start
    if "category" in hentry["properties"]:
        result["category"] = hentry["properties"]["category"]
    if "pubkey" in hentry["properties"]:
        result["pubkey"] = hentry["properties"]["pubkey"]
    if "vote" in hentry["properties"]:
        result["vote"] = hentry["properties"]["vote"]
    # NOTE patch end

    title = util.get_plain_text(hentry['properties'].get('name'))
    if title and util.is_name_a_title(title, result.get('content-plain')):
        result['name'] = title

    for prop in ('in-reply-to', 'like-of', 'repost-of', 'bookmark-of',
                 'vote-on', 'comment', 'like', 'repost'):  # NOTE added vote-on
        for url_val in hentry['properties'].get(prop, []):
            if isinstance(url_val, dict):
                result.setdefault(prop, []).append(
                    util.interpret(parsed, source_url, base_href, url_val,
                                   use_rel_syndication=False,
                                   want_json=want_json,
                                   fetch_mf2_func=fetch_mf2_func))
            else:
                result.setdefault(prop, []).append({
                    'url': url_val,
                })

    return result


util.interpret_entry = interpret_entry


def _interpret_common_properties(
        parsed, source_url, base_href, hentry, use_rel_syndication,
        want_json, fetch_mf2_func):
    result = {}
    props = hentry['properties']

    for prop in ('url', 'uid', 'photo', 'featured' 'logo'):
        value = util.get_plain_text(props.get(prop))
        if value:
            result[prop] = value

    for prop in ('start', 'end', 'published', 'updated', 'deleted'):
        date_str = util.get_plain_text(props.get(prop))
        if date_str:
            if want_json:
                result[prop] = date_str
            else:
                result[prop + '-str'] = date_str
                try:
                    date = util.parse_datetime(date_str)
                    if date:
                        result[prop] = date
                except ValueError:
                    util.logging.warn('Failed to parse datetime %s', date_str)

    author = util.find_author(parsed, source_url, hentry, fetch_mf2_func)
    if author:
        result['author'] = author

    content_prop = props.get('content')
    content_value = None
    if content_prop:
        if isinstance(content_prop[0], dict):
            content_html = content_prop[0].get('html', '').strip()
            content_value = content_prop[0].get('value', '').strip()
        else:
            content_value = content_html = content_prop[0]
        result['content'] = util.convert_relative_paths_to_absolute(
            source_url, base_href, content_html)
        result['content-plain'] = content_value

    summary_prop = props.get('summary')
    if summary_prop:
        if isinstance(summary_prop[0], dict):
            result['summary'] = summary_prop[0]['value']
        else:
            result['summary'] = summary_prop[0]

    # Collect location objects, then follow this algorithm to consolidate
    # their properties:
    # //indieweb.org/location#How_to_determine_the_location_of_a_microformat
    location_stack = [props]

    for prop in 'location', 'adr':
        vals = props.get(prop)
        if vals:
            if isinstance(vals[0], util.string_type):
                location_stack.append({'name': vals})
            else:
                location_stack.append(vals[0].get('properties', {}))

    geo = props.get('geo')
    if geo:
        if isinstance(geo[0], dict):
            location_stack.append(geo[0].get('properties', {}))
        else:
            if geo[0].startswith('geo:'):
                # a geo: URL. try to parse it.
                # //tools.ietf.org/html/rfc5870
                parts = geo[0][len('geo:'):].split(';')[0].split(',')
                if len(parts) >= 2:
                    location_stack.append({
                        'latitude': [parts[0]],
                        'longitude': [parts[1]],
                        'altitude': [parts[2]] if len(parts) >= 3 else [],
                    })

    for prop in util.LOCATION_PROPERTIES:
        for obj in location_stack:
            if obj and obj.get(prop) and not (obj == props and prop == 'name'):
                result.setdefault('location', {})[prop] = obj[prop][0]

    if use_rel_syndication:
        result['syndication'] = list(set(
            parsed.get('rels', {}).get('syndication', []) +
            hentry['properties'].get('syndication', [])))
    else:
        result['syndication'] = hentry['properties'].get('syndication', [])

    # TODO patch start
    checkin_prop = props.get('checkin')
    if checkin_prop:
        if isinstance(checkin_prop[0], dict):
            props = checkin_prop[0]['properties']
            result['checkin'] = {"name": props["name"][0]}
            try:
                result.update({"latitude": props["latitude"][0],
                               "longitude": props["longitude"][0]})
            except KeyError:
                pass
        else:
            result['checkin'] = checkin_prop[0]

    categories = props.get("category")
    if categories:
        result["category"] = categories
    # TODO patch end

    return result


util._interpret_common_properties = _interpret_common_properties



"""Braid support."""

import time

from gevent import spawn
from gevent.queue import Queue
import jsonpatch
import requests

from .framework.util import header, tx, json, JSONEncoder, Headers
from .response import OK, NoContent, Subscription

__all__ = ["publish", "subscribe", "braidify"]


def publish(url, patch_range, patch_body):
    """Publish a patch to a local or web resource using Braid."""
    if url.startswith("/"):
        tx.kv.db.publish(url, JSONEncoder().encode({"range": patch_range,
                                                    "body": patch_body}))
    else:
        requests.put(url, json={"Patches": 1})  # TODO


def subscribe(url):
    """Subscribe to a web resource using Braid."""
    while True:
        try:
            buf = []
            sections = []
            version = None
            parents = None
            patchcount = 0
            patches = []
            in_snapshot = False
            in_patchset = False
            in_patch = False
            patch_range = None
            for chunk in requests.get(url, headers={"Subscribe": "keep-alive"},
                                      stream=True, verify=False):
                buf.extend(chunk.splitlines())
                while b"" in buf:
                    delimiter = buf.index(b"")
                    sections.append(b"\n".join(buf[:delimiter]))
                    buf = buf[delimiter + 1:]
                while True:
                    try:
                        section = sections.pop(0)
                    except IndexError:
                        break
                    if in_snapshot:
                        section = section.replace(b"\n", b"")
                        yield version, parents, [(None, json.loads(section))]
                        version = None
                        in_snapshot = False
                    elif in_patchset:
                        if in_patch:
                            patchcount -= 1
                            section = section.replace(b"\n", b"")
                            patches.append((patch_range, json.loads(section)))
                            patch_range = None
                            if patchcount == 0:
                                yield version, parents, patches
                                version = None
                                patches = []
                                in_patchset = False
                                in_patch = False
                        else:
                            headers = Headers.from_lines(section.decode())
                            patch_range = headers["content-range"]
                            in_patch = True
                    else:
                        headers = Headers.from_lines(section.decode())
                        version = headers.get("version")
                        parents = headers.get("parents")
                        patchcount = int(headers.get("patches", 0))
                        if patchcount:
                            in_patchset = True
                        else:
                            in_snapshot = True
        except requests.exceptions.ChunkedEncodingError as err:
            print("---")
            print(err)
            print("connection lost. reconnecting..")
            print("---")


def multi_subscribe(path, *urls):
    """Yield patches from multiple subscriptions."""
    queue = Queue()

    def producer(url):
        def _producer():
            for patch in subscribe(url):
                queue.put_nowait(patch)
            queue.put_nowait("COMPLETED")
        return _producer

    for url in urls:
        spawn(producer(url))
    completed = 0
    while True:
        patch = queue.get(timeout=5)
        if patch == "COMPLETED":
            completed += 1
            if completed == len(urls):
                break
            continue
        tx.kv.db.publish(path, patch)
        # XXX yield patch + b"\n"


def braidify(handler, app):
    """Handle Braid Pub/Sub."""
    path = f"/{tx.request.uri.path}"
    method = tx.request.method
    headers = tx.request.headers
    controller = tx.request.controller
    if method == "OPTIONS":  # allow CORS FIXME secure with IndieAuth?
        header("Access-Control-Allow-Origin", "*")
        header("Access-Control-Allow-Methods", "GET,PUT")
        header("Access-Control-Allow-Headers",
               "credentials,subscribe,patches,client")
        header("Vary", "Origin")
        raise NoContent()
    if method == "GET":
        if "application/json" in str(tx.request.headers.get("accept")):
            try:
                resource = tx.db.select("resources", where="path = ?",
                                        vals=[path])[0]
            except IndexError:
                pass  # not a braided resource
            else:
                raise OK(JSONEncoder().encode(dict(resource["resource"])))
        elif headers.get("Subscribe") == "keep-alive":
            header("Access-Control-Allow-Origin", "*")
            header("Subscribe", "keep-alive")
            header("Content-Type", "application/json")
            header("X-Accel-Buffering", "no")
            try:
                subscription = controller._subscribe()
            except AttributeError:
                subscription = Braid(path)
            tx.response.naked = True
            raise Subscription(subscription)
    if method == "PUT" and headers.get("Patches"):
        # TODO handle multiple patches
        print(tx.request.body)
        patch_range, patch_body = parse_patch(tx.request.body)
        resource = tx.db.select("resources", where="url = ?",
                                vals=[path])[0]["resource"]
        if patch_range.endswith("-"):  # TODO FIXME
            op = "add"
        else:
            op = "replace"
        patch = {"op": op, "path": patch_range,
                 "value": patch_body}
        jsonpatch.apply_patch(resource, [patch], in_place=True)
        tx.db.update("resources", resource=resource,
                     where="url = ?", vals=[path])
        publish(path, patch_range, json.loads(patch_body))
        header("Access-Control-Allow-Origin", "*")
        header("Patches", "OK")
        raise OK(b"")
    yield


def parse_patch(patch, delimiter):
    """Parse a single patch from a GET or a PUT."""
    raw_headers, _, patch_body = patch.partition(delimiter)
    patch_headers = Headers.from_lines(raw_headers.decode())
    # TODO version = patch_headers.get("version")
    # TODO parents = patch_headers.get("parents")
    # TODO merge_type = patch_headers.get("merge-type")
    try:
        patch_range = str(patch_headers["content-range"]).partition("=")[2]
    except KeyError:
        patch_range = None  # snapshot
    print(patch_body)
    return patch_range, json.loads(patch_body)


class Braid:
    """A Braid subscription."""

    def __init__(self, path):
        """Create Redis subscription to resource at given path."""
        self.p = tx.kv.pubsub(ignore_subscribe_messages=True)
        self.p.subscribe(path)

    def __next__(self):
        """Serve patches to client when received from Redis subscription."""
        while True:
            message = self.p.get_message()
            if message is None:
                time.sleep(0.001)
                continue
            patch = json.loads(message["data"])
            body = JSONEncoder().encode(patch["body"])
            return bytes(f"Content-Length: {len(body)}\n"
                         f"Content-Range: json={patch['range']}\n"
                         f"\n"
                         f"{body}\n"
                         f"\n", "utf-8")

    def __iter__(self):  # TODO XXX use iter() on Braid object above?
        """Identify and function as an iterator."""
        return self



"""

"""

from . import util

__all__ = ["CacheControl", "Connection", "Date", "Pragma", "Trailer",
           "TransferEncoding", "Upgrade", "Via", "Warning"]


class General(util.Header):

    """"""


class CacheControl(General):

    """"""


class Connection(General):

    """"""


class Date(General):

    """"""


class Pragma(General):

    """"""


class Trailer(General):

    """"""


class TransferEncoding(General):

    """"""


class Upgrade(General):

    """"""


class Via(General):

    """"""


class Warning(General):

    """"""



"""
[Public Suffix List][1] support

    >>> split("www.example.org")
    'www', 'example', 'org'

Based upon the original implementation [`publicsuffix`][2] copyright
@[Tomaž Solc][3] and released under an MIT license.

[1]: //publicsuffix.org/list/
[2]: //pypi.python.org/pypi/publicsuffix
[3]: //tablix.org

"""

import codecs
import collections
import pkg_resources

import requests


_suffixes = None
Domain = collections.namedtuple("Domain", "subdomain domain suffix")


def split(hostname):
    """
    return the subdomain and domain of given `hostname`

        >>> split("www.example.org")
        'www', 'example', 'org'
        >>> split("www.example.org.uk")
        'www', 'example', 'org.uk'

    """
    # TODO handle Punycode decoding
    if _suffixes is None:
        _initialize()
    parts = hostname.lower().lstrip(".").split(".")
    hits = [None] * len(parts)
    _suffixes.lookup(hits, 1, parts)
    for i, what in enumerate(hits):
        if what is not None and what == 0:
            return Domain(".".join(parts[:i]), parts[i], ".".join(parts[i+1:]))


def _initialize():
    global _suffixes
    _suffixes = _PublicSuffixList()


class _PublicSuffixList:

    """
    reads and parses the public suffix list

    """

    def __init__(self):
        input_path = pkg_resources.resource_filename("understory.uri",
                                                     "public_suffix_list.dat")
        try:
            with codecs.open(input_path, "r", "utf8") as fp:
                self._build_structure(fp)
        except FileNotFoundError:
            res = requests.get("https://publicsuffix.org/list/"
                               "public_suffix_list.dat")
            with codecs.open(input_path, "w", "utf8") as fp:
                fp.write(res.text)
            with codecs.open(input_path, "r", "utf8") as fp:
                self._build_structure(fp)

    def lookup(self, matches, depth, parts, parent=None):
        if parent is None:
            parent = self.root
        if parent in (0, 1):
            negate = parent
            children = None
        else:
            negate, children = parent
        matches[-depth] = negate
        if depth < len(parts) and children:
            for name in ("*", parts[-depth]):
                child = children.get(name, None)
                if child is not None:
                    self.lookup(matches, depth+1, parts, child)

    def _build_structure(self, fp):
        root = [0]
        for line in fp:
            line = line.strip()
            if line.startswith("//") or not line:
                continue
            self._add_rule(root, line.split()[0].lstrip("."))
        self.root = self._simplify(root)

    def _add_rule(self, root, rule):
        if rule.startswith("!"):
            negate = 1
            rule = rule[1:]
        else:
            negate = 0
        parts = rule.split(".")
        self._find_node(root, parts)[0] = negate

    def _find_node(self, parent, parts):
        if not parts:
            return parent
        if len(parent) == 1:
            parent.append({})
        assert len(parent) == 2
        negate, children = parent
        child = parts.pop()
        child_node = children.get(child, None)
        if not child_node:
            children[child] = child_node = [0]
        return self._find_node(child_node, parts)

    def _simplify(self, node):
        if len(node) == 1:
            return node[0]
        return (node[0], dict((k, self._simplify(v)) for
                              (k, v) in node[1].items()))



"""Mountable IndieWeb apps and helper functions."""

from ... import mm
from .. import framework as fw
from . import indieauth
from . import micropub
from . import microsub
from . import webmention
from . import websub

__all__ = ["indieauth", "micropub", "microsub", "webmention", "websub",
           "cache_app"]


cache_app = fw.application("Cache", db=False, mount_prefix="admin/cache",
                           resource=r".+")
tmpl = mm.templates(__name__)


@cache_app.route(r"")
class Cache:

    def get(self):
        return tmpl.cache(fw.tx.db.select("cache"))


@cache_app.route(r"{resource}")
class Resource:

    def get(self):
        resource = fw.tx.db.select("cache", where="url = ? OR url = ?",
                                   vals=[f"https://{self.resource}",
                                         f"http://{self.resource}"])[0]
        return tmpl.cache_resource(resource)



"""WebSub hub, bub."""

from understory import web
from understory.web import tx


hub = web.application("WebSubHub", mount_prefix="hub", db=False,
                      subscription_id=r".+")
templates = web.templates(__name__)
subscription_lease = 60 * 60 * 24 * 90


def publish(topic_url, callback_url):
    """"""
    callback_url
    for channel in topic_url:
        pass
    web.post(callback_url)


def subscribe(url, callback_url):
    """Send subscription request."""
    topic_url = web.discover_link(url, "self")
    hub = web.discover_link(url, "hub")
    web.post(hub, data={"hub.mode": "subscribe", "hub.topic": str(topic_url),
                        "hub.callback": callback_url})


def verify_subscription(topic_url, callback_url):
    """Verify subscription request and add follower to database."""
    verification_data = {"hub.mode": "subscribe", "hub.topic": topic_url,
                         "hub.challenge": web.nbrandom(32),
                         "hub.lease_seconds": subscription_lease}
    response = web.get(callback_url, params=verification_data)
    if response.text != verification_data["hub.challenge"]:
        return
    tx.db.insert("followers", topic_url=web.uri(topic_url).path,
                 callback_url=str(web.uri(callback_url)))


def wrap(handler, app):
    """Ensure server links are in head of root document."""
    tx.db.define(followers="""followed DATETIME NOT NULL DEFAULT
                                  CURRENT_TIMESTAMP,
                              topic_url TEXT, callback_url TEXT""")
    yield
    # TODO limit to subscribables
    if tx.request.uri.path == "" and tx.response.body:
        doc = web.parse(tx.response.body)
        try:
            head = doc.select("head")[0]
        except IndexError:
            pass
        else:
            head.append(f"<link rel=self href=/{tx.request.uri.path}>")
            head.append("<link rel=hub href=/hub>")
            tx.response.body = doc.html
        web.header("Link", f'/<{tx.request.uri.path}>; rel="self"', add=True)
        web.header("Link", '</hub>; rel="hub"', add=True)


@hub.route(r"")
class Hub:
    """."""

    def get(self):
        return templates.hub(tx.db.select("followers"))

    def post(self):
        mode = web.form("hub.mode")["hub.mode"]
        if mode != "subscribe":
            raise web.BadRequest("hub only supports subscription; "
                                 '`hub.mode` must be "subscribe"')
        form = web.form("hub.topic", "hub.callback")
        # TODO raise web.BadRequest("topic not found")
        web.enqueue(verify_subscription, form["hub.topic"],
                    form["hub.callback"])
        raise web.Accepted("subscription request accepted")


@hub.route(r"{subscription_id}")
class Subscription:
    """."""

    def get(self):
        """Confirm subscription request."""
        form = web.form("hub.mode", "hub.topic", "hub.challenge",
                        "hub.lease_seconds")
        # TODO verify the subscription
        return form["hub.challenge"]

    def post(self):
        """Check feed for updates."""
        print(tx.request.body._data)
        return



import term

import mm


main = term.application("mm", __doc__)


@main.register()
class Main:

    def setup(self, parser):
        add = parser.add_argument
        add("args", nargs="*", help="argument(s) passed to template")
        add("-w", "--wrap", nargs="+", dest="wrappers", type=open,
            help="wrap in given template(s)")

    def run(self, args, stdin):
        document = mm.Template(stdin)(*args.args)
        for wrapper in args.wrappers:
            document = mm.Template(wrapper)(document)
        print(document)
        return 0



"""

"""

from .util import Status

__all__ = ["Informational", "Continue", "SwitchingProtocols"]


class Informational(Status):

    """1xx -- request received, continuing process"""


class Continue(Informational):

    """
    100 [Part II S 8.1.1][0]

    The client SHOULD continue with its request.This interim response is
    used to inform the client that the initial part of the request has
    been received and has not yet been rejected by the server. The client
    SHOULD continue by sending the remainder of the request or, if the
    request has already been completed, ignore this response. The server
    MUST send a final response after the request has been completed. See
    Section 7.2.3 of [Part1] for detailed discussion of the use and
    handling of this status code.

    [0]: http://tools.ietf.org/html/httpbis-p2-semantics#section-8.1.1

    """


class SwitchingProtocols(Informational):

    """101"""



"""
An intelligent Pythonic URI interface.

    >>> uri = parse("https://en.wikipedia.org/wiki/Jabberwocky")
    >>> uri.is_secure
    True
    >>> uri.scheme
    'https'
    >>> uri.host, uri.port
    'en.wipedia.org', 80
    >>> uri.subdomain, uri.domain
    'en', 'wikipedia.org'
    >>> uri.path
    '/wiki/Jabberwocky'

"""

import base64
import hashlib
import inspect
import mimetypes
import unicodedata
import urllib.error
import urllib.parse
import urllib.request

import hstspreload

from . import publicsuffix

__all__ = ["parse", "supported_schemes", "HTTPURI", "HTTPSURI", "WSURI",
           "WSSURI", "DataURI", "JavascriptURI", "MagnetURI"]


def parse(uri):  # XXX , secure=True):
    """Return a `URI` object for given `uri`.

    Various web-related protocols supported.

        >>> webpage = parse("https://tantek.com")
        >>> stylesheet = parse("data:text/css,body{font:14px/1.5 Helvetica;}")
        >>> script = parse("javascript:alert('hello world')")

    """
    # TODO FIXME XXX NOTES: simple and expected...
    uri = str(uri)
    if ":" in uri:
        scheme, _, identifier = uri.partition(":")
    else:
        scheme = "http"

    # XXX first_slash = scheme.find("/")
    # XXX if not identifier or (first_slash > 0 and len(scheme) > first_slash):
    # XXX     scheme = "http"
    # XXX     identifier = uri

    try:
        handler = supported_schemes[scheme]
    except KeyError:
        raise ValueError(f"scheme `{scheme}` not supported")

    return handler(uri)

    # XXX uri = handler(identifier)
    # XXX # TODO if scheme == "https":  # TODO cleanup
    # XXX if secure:
    # XXX     uri.is_secure = True
    # XXX     uri.scheme = "https"
    # XXX elif isinstance(uri, HTTPSURI):
    # XXX     uri.is_secure = True
    # XXX     uri.scheme = "https"
    # XXX     if uri.suffix == "onion":
    # XXX         uri.is_secure = False
    # XXX         uri.scheme = "http"
    # XXX return uri


def clean(s):
    # XXX s = str(urllib.parse.unquote(s), "utf-8", "replace")
    s = urllib.parse.unquote(s)
    return unicodedata.normalize("NFC", s).encode("utf-8")


class URI:

    """
    a Uniform Resource Identifier

    """

    def __init__(self, uri):
        self.given = uri

    def __eq__(self, other):
        return str(self) == str(parse(other))

    def __ne__(self, other):
        return not self.__eq__(other)

    def __len__(self):
        return len(self.normalized)

    def __str__(self):
        return self.normalized

    def __bytes__(self):
        return bytes(self.normalized, "utf-8")

    def __add__(self, suffix):
        new = "".join((self.given, str(suffix)))
        return self.__class__(new)

    def __truediv__(self, path):
        new = "/".join((self.given, str(path)))
        return self.__class__(new)

    def __hash__(self):
        try:
            return self.__hash
        except AttributeError:
            pass
        self.__hash = int(hashlib.sha1(bytes(self.normalized,
                                             "utf-8")).hexdigest(), 16)
        return self.__hash

    def __repr__(self):
        return self.normalized


class HTTPURI(URI):

    """
    web address

        >>> HTTPURI("//example.org")
        http://example.org

    """

    is_secure = False

    def __init__(self, identifier):
        super().__init__(identifier)
        self._normalize()

    def is_hsts(self):
        return hstspreload.is_hsts_preload(self.host)

    @classmethod
    def from_parts(cls, netloc, path="/", query=None, fragment=""):
        """
        instantiate a URI from parts

            >>> HTTPURI.from_parts("example.org")
            http://example.org

        """
        if query is None:
            query = {}
        query_string = urllib.parse.urlencode(query, doseq=True)
        return cls(urllib.parse.urlunsplit((cls.__name__.lower(), netloc,
                                            path, query_string, fragment)))

    @property
    def dict_items(self):
        return dict(scheme=self.scheme, host=self.host, path=self.path)

    @property
    def minimized(self):
        uri = self.normalized
        uri = uri[len(self.scheme) + 3:]
        if uri.startswith("www."):
            uri = uri[4:]
        # FIXME strip trail slash on path not fragment
        return uri.rstrip("/").partition("#")[0]

    @property
    def normalized(self):
        query = urllib.parse.urlencode(self.query, doseq=True)
        normalized_parts = ("https" if self.is_secure else "http",
                            self.netloc, self.path, query, self.fragment)
        normalized = urllib.parse.urlunsplit(normalized_parts)
        # XXX if self.is_relative:
        # XXX     normalized = "/" + normalized
        return normalized

    def _normalize(self):
        uri = self.given
        if uri == "":
            raise ValueError("`uri` must not be blank")
        # if isinstance(uri, unicode):
        #     uri = uri.encode("utf-8", "ignore")

        if uri.startswith("//"):
            self.is_absolute = True
            uri = uri[2:]
        if not uri.startswith(("/", "http://", "https://")):
            uri = "http://" + uri

        uri = uri.replace("#!", "?_escaped_fragment_=", 1)

        parts = urllib.parse.urlsplit(uri.strip())
        self.netloc = ""
        if parts.scheme:
            self.scheme = self._normalize_scheme(parts.scheme)
            self.username = self._normalize_username(parts.username)
            self.password = self._normalize_password(parts.password)
            self.host = self._normalize_host(parts.hostname)
            self.port = self._normalize_port(parts.port)
            if self.username:
                auth = self.username
                if self.password:
                    auth += ":" + self.password
                self.netloc = auth + "@"
            # if not self.host:
            #     raise ValueError("no host in an absolute `uri`")
            self.netloc += self.host
            if self.port != 80:
                self.netloc += ":" + str(self.port)
            self.subdomain, self.domain, self.suffix = \
                publicsuffix.split(parts.hostname)
        self.path = self._normalize_path(parts.path).lstrip("/")
        self.raw_query = self._normalize_query(parts.query)
        self.query = urllib.parse.parse_qs(self.raw_query)
        self.fragment = self._normalize_fragment(parts.fragment)

    def _normalize_scheme(self, scheme):
        if scheme not in ("http", "https"):
            error_msg = "`{0}` scheme not supported".format(scheme)
            raise ValueError(error_msg)
        scheme = scheme.lower()
        return scheme

    def _normalize_username(self, username):
        if username is None:
            username = ""
        return username

    def _normalize_password(self, password):
        if password is None:
            password = ""
        return password

    def _normalize_host(self, host):
        if host is None:
            raise ValueError("absolute `uri` requires a host")
        if " " in host:
            raise ValueError("spaces not allowed in host")
        # host = host.lower().strip(".").decode("utf-8").encode("idna")
        host = host.lower().strip(".")  # .encode("idna")
        return host

    def _normalize_port(self, port):
        # TODO limit to range of possibilities (0 < port < 36???)
        if port is None:
            port = 80
        return port

    def _normalize_path(self, path):
        if path == "":
            path = "/"
        path = urllib.parse.unquote(path)
        path = urllib.parse.quote(path, "~:/?#[]@!$&'()*+,;=")
        # path = self._clean(path)
        # XXX if self.is_absolute:
        output = []
        for part in path.split("/"):
            if part == "":
                if not output:
                    output.append(part)
            elif part == ".":
                pass
            elif part == "..":
                if len(output) > 1:
                    output.pop()
            else:
                output.append(part)
        if part in ["", ".", ".."]:
            output.append("")
        path = "/".join(output)
        return path

    def _normalize_query(self, query):
        # TODO %3a to %3A
        # TODO %7E to ~
        args = ["=".join([urllib.parse.quote(clean(t), "~:/?#[]@!$'()*+,;=")
                for t in q.split("=", 1)]) for q in query.split("&")]
        return "&".join(args)

    def _normalize_fragment(self, fragment):
        fragment = urllib.parse.unquote(fragment)
        fragment = urllib.parse.quote(fragment, "~")
        return fragment

    def __getitem__(self, key):
        """get a query parameter"""
        try:
            return self.query[key]
        except KeyError:
            self.query[key] = []
            return self.query[key]

    def __setitem__(self, key, value):
        """set a query parameter"""
        if isinstance(value, list):
            self.query[key] = value
        else:
            self.query[key] = [value]

    def update(self, **args):
        self.query.update(**args)


class HTTPSURI(HTTPURI):

    """
    secure web address

        >>> HTTPSURI("//example.org")
        https://example.org

    """

    is_secure = True

    def __init__(self, identifier):
        super().__init__(identifier)


class WSURI(URI):

    """
    WebSocket service endpoint

        >>> WSURI("//example.org")
        ws://example.org

    """

    is_secure = False


class WSSURI(WSURI):

    """
    secure WebSocket service endpoint

        >>> WSSURI("//example.org")
        wss://example.org

    """

    is_secure = True

    def __init__(self, identifier):
        super().__init__(identifier)


class DataURI(URI):

    """
    data objects

        >>> DataURI("foo bar")
        data:,foo bar
        >>> data = DataURI("iVBORw0KGgoAAAANSUhEUgAAAAUAAAAFCAYAAACNbyblAAAAH"
        ...                "ElEQVQI12P4//8/w38GIAXDIBKE0DHxgljNBAA09O9TXL0Y4O"
        ...                "HwAAAABJRU5ErkJggg==",
        ...                encoded=True, mime_type="image/png")
        >>> data  # doctest: +ELLIPSIS
        data:image/png;base64,iVBORw0K...rkJggg==
        >>> data.mime_type
        "image/png"
        >>> data.save("foo.png")  # doctest: +SKIP

    """

    def __init__(self, data, encoded=False,
                 mime_type="text/plain", charset="US-ASCII"):
        self.data = data
        self.encoded = encoded
        self.mime_type = mime_type
        self.charset = charset

    @classmethod
    def from_identifier(cls, identifier):
        r"""
        return a data URI for given parsed identifier

            >>> Data.from_identifier("text/html,<!doctype html>foo bar")
            data:text/html,<!doctype html>foo bar
            >>> Data.from_identifier("text/html;charset=utf-8"
            ...                      ",<!doctype html>fnṏrd")
            data:text/html;charset=utf-8,foo
            >>> Data.from_identifier("charset=utf-8,\xe2\x81\x82")
            data:charset=utf-8,foo

        """
        metadata, _, data = identifier.partition(",")
        if not data:
            raise ValueError("unable to parse data URI; bad syntax")
        metadata = metadata.lower().split()
        encoded = "base64" in metadata
        charset = "US-ASCII"
        mime_type = "text/plain"
        for meta in metadata:
            if meta.startswith("charset="):
                charset = meta.partition("=")[2]
            elif meta != "base64":
                mime_type = meta
        return cls(data, encoded, mime_type, charset)

    @classmethod
    def from_file(cls, path, mime_type=None):
        """
        return a data URI for contents of file at given path

        MIME type will be inferred from the file extension if possible. You
        may override this by providing your own with `mime_type`.

            >>> Data.from_file("glider.png")  # doctest: +SKIP
            data:image/png;base64,...

        """
        # TODO infer charset
        with open(path, "rb") as f:
            data = f.read()
        mime_type = mimetypes.guess_type(path)[0]
        encoded = False
        if mime_type.startswith(("image", "audio", "video")):
            data = base64.b64encode(data)
            encoded = True
        return cls(data, mime_type, encoded=encoded)

    @property
    def normalized(self):
        metadata = []
        if self.mime_type != "text/plain":
            metadata.append(self.mime_type)
        if self.charset != "US-ASCII":
            metadata.append("charset={}".format(self.charset))
        if self.encoded:
            metadata.append("base64")
        return "data:{},{}".format(";".join(metadata), self.data)


class JavascriptURI(URI):

    """
    JavaScript code

        javascript:<javascript to execute>

        >>> JavascriptURI("alert('example');")
        javascript:alert('example');

    """

    @classmethod
    def from_identifier(cls, identifier):
        """

        """
        print(identifier)
        return cls()

    @property
    def normalized(self):
        return "javascript:{}".format(self.given)


class MagnetURI(URI):

    """
    address to a specific piece of content

        magnet:<content-parameters>

        >>> MagnetURI("")
        magnet:alert('example');

    """

    def __init__(self, identifier):
        super().__init__(identifier)
        self._normalized = urllib.parse.unquote(identifier)

    @property
    def normalized(self):
        return "magnet:{}".format(self._normalized)


class TelURI(URI):

    """"""

    def __init__(self, identifier):
        super().__init__(identifier)

    @property
    def normalized(self):
        return "tel:{}".format(self._normalized)


class FaxURI(URI):

    """"""

    def __init__(self, identifier):
        super().__init__(identifier)

    @property
    def normalized(self):
        return "fax:{}".format(self._normalized)


class SMSURI(URI):

    """"""

    def __init__(self, identifier):
        super().__init__(identifier)
        self._normalize()

    def _normalize(self):
        uri = self.given
        parts = urllib.parse.urlsplit(uri.strip())
        self.numbers = parts.path.split(",")
        self.body = urllib.parse.parse_qs(parts.query).get("body", [None])[0]
        self._normalized = ",".join(self.numbers)
        if self.body:
            self._normalized += "?body={}".format(parts.query)

    @property
    def normalized(self):
        return "sms:{}".format(self._normalized)


class WEB_ACTIONURI(URI):

    """"""

    def __init__(self, identifier):
        super().__init__(identifier)
        self._normalize()

    def _normalize(self):
        uri = self.given
        parts = urllib.parse.urlsplit(uri.strip())
        self.action = parts.netloc
        self.query = urllib.parse.parse_qs(parts.query)
        self._normalized = f"{parts.netloc}?{parts.query}"

    @property
    def normalized(self):
        return f"web+action://{self._normalized}"

    def __getitem__(self, key):
        """get a query parameter"""
        return self.query[key]


class MOZ_EXTENSIONURI(URI):

    """"""

    def __init__(self, identifier):
        super().__init__(identifier)
        self._normalize()

    def _normalize(self):
        uri = self.given
        parts = urllib.parse.urlsplit(uri.strip())
        self.parts = parts
        self.host = parts.netloc
        self.query = urllib.parse.parse_qs(parts.query)
        self._normalized = f"{parts.netloc}{parts.path}"
        if self.query:
            self._normalized += f"?{parts.query}"

    @property
    def normalized(self):
        return f"moz-extension://{self._normalized}"

    def __getitem__(self, key):
        """get a query parameter"""
        return self.query[key]

    def __setitem__(self, key, value):
        """set a query parameter"""
        if isinstance(value, list):
            self.query[key] = value
        else:
            self.query[key] = [value]


supported_schemes = {}
for scheme, obj in dict(globals()).items():
    if inspect.isclass(obj) and issubclass(obj, URI):
        scheme = scheme.lower()[:-3]
        separator = "-"
        if scheme.startswith("web"):
            separator = "+"
        supported_schemes[scheme.replace("_", separator)] = obj
